/*
	Header inclusions.
*/

#include <WebCalc.h>



/*
	DomStyle manipulation.
*/

/**
 *	DomStyle_new
 *
 *	Dynamically allocates, initializes, and returns a new DomStyle object.
 *	Hexadecimal notation in default value indicate properties that should
 *	be overwritten by the tag nature.
 *
 *	@return		Allocated and initialized object.
 */

struct DomStyle *DomStyle_new(void)
{
	// Storing the size of a structure.
	const int size = sizeof(struct DomStyle);

	// Using a DomStyle structure pointer and a static default structure.
	struct DomStyle *pointer = NULL;
	struct DomStyle style = {
		// White background and black font color.
		0, 1,
		// Automatic overflow on both axes and no resize.
		3, 3, 0,
		// No display, static positioning, no floating, and top
		// vertical alignment.
		0x0, 0, 0, 0,
		// Position : top, right, bottom and left.
		0x0, 0x0, 0x0, 0x0,
		// Default width and width : value, minimum and maximum.
		0x0, 0, 4095, 0x0, 0, 4095,
		// Default padding, borders with border radius, and margins.
		0, 0, 0, 0,
		0, 0, 0, 0, 0,
		0x0, 0, 0x0, 0,
		// Border spacing on x and y.
		0, 0,
		// Font size, letter spacing, line height and text indent.
		0x0, 1, 0x0, 0,
		// Default cursor, visible and `o` list type.
		1, 1, 7,
		// Left text alignment, no decoration, default wrapping
		// behavior, separate table borders and show empty cells.
		0, 0, 0, 0, 0,
		// Tab size and word spacing.
		4, 0x0
	};

	// Allocating the needed size in the memory.
	pointer = (struct DomStyle *)malloc(size);
	// Returning a NULL pointer on alloc failure.
	if(!pointer) return NULL;

	// Copying the static data to the reserved location.
	memcpy(pointer, &style, size);
	// Returning the allocated pointer.
	return pointer;
}



/*
	DomAttribute manipulation.
*/

/**
 *	DomAttribute_new()
 *
 *	Allocates a new attribute with the given parameters. If a null
 *	reference is given for at least name or value, nothing is allocated
 *	and an empty structure is returned.
 */

struct DomAttribute *DomAttribute_new(reference_t name, reference_t value)
{
	// Using a DomAttribute structure pointer.
	struct DomAttribute *at = NULL;

	// Allocating the size of the structure in the memory.
	at = (struct DomAttribute *)malloc(sizeof(struct DomAttribute));
	// Returning a NULL pointer on alloc failure.
	if(!at) return NULL;

	// Returning an empty structure if one of the arguments is null.
	if(refnull(name) || refnull(value))
	{
		// Setting NULL pointers.
		at->name = NULL;
		at->value = NULL;
		// Returning the structure.
		return at;
	}

	// If given arguments are valid, extracting the referenced data.
	at->name = refxtrct(name);
	at->value = refxtrct(value);
	// Returning the obtained structure. No alloc failure handling is added
	// because it could not do anything more that returning the pointer.
	return at;
}

/**
 *	DomAttribute_free()
 *
 *	Frees the memory allocated for a DomAttribute object. As a convention,
 *	the structure itself is freed.
 *
 *	@param	at	Attribute to free.
 */

void DomAttribute_free(struct DomAttribute *at)
{
	// Returning is the argument is not valid.
	if(!at) return;

	// Freeing the name and value pointers, if valid.
	if(at->name) free(at->name);
	if(at->value) free(at->value);

	// Freeing the structure itself.
	free(at);
}



/*
	DomElement object
*/

/**
 *	DomElement_new()
 *
 *	Allocates and constructs a new DomElement, using the given parameters.
 *
 *	@param	name		Element tag name.
 *	@param	args		Argument string.
 *	@param	parent		Parent node in DOM tree.
 *	@param	previous	Previous sibling in DOM tree.
 *
 *	@return		Allocated object, or NULL on failure.
 */

struct DomElement *DomElement_new(reference_t name, reference_t args,
	struct DomElement *parent, struct DomElement *previous)
{
	// Using a DomElement pointer.
	struct DomElement *el;
	// Using a DomStyle pointer to test allocation validity.
	struct DomStyle *style;

	// Allocating the pointer.
	el = (struct DomElement *)malloc(sizeof(struct DomElement));
	// Handling alloc failure by returning NULL without doing anything.
	if(!el) return NULL;

	// Immediately getting a style structure (to return on failure).
	style = DomStyle_new();
	// Handling alloc failure by canceling the operation.
	if(!style)
	{
		// Freeing the allocated element.
		free(el);
		// Returning from the function.
		return NULL;
	}
	// Setting the style structure pointer.
	el->style = style;

	// Extracting the tag name.
	el->tagName = refxtrct(name);

	// Setting a default textNode (or tex, basically even with the
	// uncertainty on union manipulation this should be the same).
	el->textNode = NULL;

	// Setting default values for children.
	el->children = NULL;
	el->childnum = 0;

	// Setting default values for attributes.
	el->attributes = NULL;
	el->attrnum = 0;

	// Setting the parent and previous sibling.
	el->parent = parent;
	el->previous = previous;

	// Returning the allocated structure.
	return el;
}

/**
 *	DomElement_free()
 *
 *	Frees a DomElement structure and all the associated allocate data. All
 *	its children are freed. As a convention, the structure itself is also
 *	freed.
 *
 *	@param	el	Element structure to free.
 */

void DomElement_free(struct DomElement *el)
{
	// Using an iterator.
	int i;

	// Freeing the tex node data.
	if(!strcmp(el->tagName, "math")) tex_free(el->tex);
	// Freeing the node content.
	else if(el->textNode) free(el->textNode);

	// Freeing the tag name, if not NULL.
	if(el->tagName) free(el->tagName);

	// Freeing the style structure.
	free(el->style);

	// Freeing the tag attributes and the array.
	for(i=0; i < el->attrnum; i++) DomAttribute_free(el->attributes[i]);
	if(el->attributes) free(el->attributes);

	// Freeing the children and the array.
	for(i=0; i < el->childnum; i++) DomElement_free(el->children[i]);
	if(el->children) free(el->children);

	// Finally freeing the structure itself.
	free(el);
}

/**
 *	DomElement_addChild()
 *
 *	Adds a child to an element. Needs to reallocate the children array.
 *
 *	@param	el	Parent element.
 *	@param	child	Child to add.
 *
 *	@return		Argument `child`, or NULL on failure.
 */

struct DomElement *DomElement_addChild(struct DomElement *el,
	struct DomElement *child)
{
	// Using a pointer array pointer.
	struct DomElement **array;
	// Using a variable to store the number of children.
	int n;

	// Checking arguments validity. Returning NULL on error.
	if(!el || !child) return NULL;

	// Getting the number of children.
	n = el->childnum;

	// Allocating a new array, freeing the old one on success.
	array = realloc(el->children, (n + 1) * sizeof(struct DomElement *));
	// Canceling operation on failure.
	if(!array) return NULL;

	// Adding the new child.
	array[n] = child;

	// Updating parent information.
	el->children = array;
	el->childnum++;

	// Returning the child element.
	return child;
}

/**
 *	DomElement_getChild()
 *
 *	Returns the pointer on an element's child, represented by its position
 *	in the parent child list.
 *
 *	@param	el	Element whose child is wanted.
 *	@param	n	Number of wanted child.
 *
 *	@return		Pointer of child element if exists, NULL otherwise.
 */

struct DomElement *DomElement_getChild(struct DomElement *el, int n)
{
	// Using an integer to store the number of children.
	int children;

	// Checking argument validity.
	if(!el) return NULL;

	// Getting the number of children.
	children = el->childnum;

	// Adapting range of n.
	n %= children;
	if(n < 0) n += children;

	// Returning the wanted child.
	return el->children[n];
}

/**
 *	DOM_sybl()
 */

static int DOM_sybl(reference_t name)
{
	const char *names[] = { "amp", "lt", "gt", NULL };
	const char symbols[] = { '&', '<', '>' };
	const char **ptr = names;

	while(*ptr) if(!refscmp(name, *ptr++)) return symbols[ptr - names - 1];
	return ' ';
}

/**
 *	DomElement_setTextNode()
 *
 *	Sets the text node of an element. If a text node already exists, it is
 *	freed. To remove the text node of an element, just call setTextNode(
 *	NULL).
 *
 *	@param	el	Element whose text node should be set.
 *	@param	text	Text content.
 */

void DomElement_setTextNode(struct DomElement *el, reference_t text)
{
	char *txt;
	const char *ptr;
	int length = 0, i = 0;

	ptr = text.begin;
	while(ptr < text.end)
	{
		if(*ptr == '&')
		{
			while(ptr < text.end && *ptr != ';') ptr++;
			if(ptr == text.end) break;
		}

		length++;
		ptr++;
	}

	txt = malloc(length + 1);

	ptr = text.begin;
	while(ptr < text.end)
	{
		if(*ptr == '&')
		{
			const char *end = ptr;
			reference_t sybl = { ptr + 1, ptr };

			while(end < text.end && *end != ';') end++;
			sybl.end = end;

			*(txt + i++) = DOM_sybl(sybl);

			if(end == text.end) break;
			ptr = end + 1;
			continue;
		}

		*(txt + i++) = *ptr++;
	}

	*(txt + i) = 0;

	if(!strcmp(el->tagName, "math"))
	{
		el->tex = tex_load_pointer(txt, txt + i);
		free(txt);
		return;
	}

	el->textNode = txt;
}

/**
 *	DomElement_getAttribute()
 *
 *	Returns the value associated to a given attribute name.
 *
 *	@param	el	Concerned element.
 *	@param	name	Attribute name.
 *
 *	@return		Attribute value if exists, NULL otherwise.
 */

const char *DomElement_getAttribute(struct DomElement *el, const char *name)
{
	// Using an attribute pointer.
	struct DomAttribute *attribute;
	// Using an iterator.
	int i;

	// Checking arguments validity.
	if(!el || !name) return NULL;

	// Seeking the wanted attribute.
	for(i=0; i < el->attrnum; i++)
	{
		// Getting the current attribute.
		attribute = el->attributes[i];

		// Returning the value if the name corresponds.
		if(!strcmp(attribute->name, name)) return attribute->value;
	}

	// If nothing is found, returning NULL;
	return NULL;
}

/**
 *	DomElement_resolve()
 *
 *	Recursively computes width and height of an element and all its
 *	children.
 *	Top and left coordinates indicate the position where the element is
 *	placed, including padding, borders and margins.
 *	Basic information (position, maximum width, etc.) of the current
 *	element should be set by the parent call of DomElement_resolve(). As
 *	a consequence, it is advised to use only this function on the html
 *	element, that is, by calling document_resolve().
 */

void DomElement_resolve(struct DomElement *el)
{
	// Using a DomElement structure pointer for iterating over children.
	struct DomElement *child;
	// Using two variables for coordinates handling and an iterator.
	int x, y, i;
	// Using an integer to store the maximum width of children.
	int mw;

	// Resolving math elements.
	if(!strcmp(el->tagName,"math"))
	{
		// Setting the dimensions given by the natural display module.
		el->style->width = el->tex->width;
		el->style->height = el->tex->height;
	}
	// Resolving other text elements.
	else if(el->textNode) Disp_text(el,1,0,0);

	// Applying special behavior for each display style.
	switch(el->style->display)
	{
		// Block elements.
		case 0:
			el->style->width = el->style->maxWidth;
	}

	// Returning immediately if the element hasn't any child.
	if(!el->childnum) return;

	// Resolving childrens.
	for(i=0; i < el->childnum; i++)
	{
		// Getting the current child.
		child = el->children[i];

		// Calculating child's maximum width.
		mw = el->style->maxWidth - DOMElement_tokenWidth(child);

		// Setting its maximum width.
		if(!child->style->maxWidth || child->style->maxWidth > mw)
			child->style->maxWidth = mw;

		// Resolving it. Then, width and height should exist.
		DomElement_resolve(child);
	}

	// Initializing cursor position.
	x = 0, y = 0;

	// Positioning children.
	for(i=0;i<el->childnum;i++)
	{
		// Getting the next child.
		child = el->children[i];

		// If next child is a block.
		if(child->style->display == 0)
		{
			// Add child's total height.
			child->style->left = x;
			child->style->top = y;

			y += DOMElement_totalHeight(child);
			x = 0;
		}
	}

	// Getting new information.
	el->style->height = y;
}

//#############################################################################################################################
//
//                                         D O M E l e m e n t   O b j e c t
//
//#############################################################################################################################

int DomElement_get(struct DomElement *el, int n)
{
	/*
	**  This function returns various informations about an element.
	**   n   return value
	**   1   total width
	**   2   total height
	**   3   border, padding, margin width
	**   4   border, padding, margin height
	*/

	switch(n)
	{
		case 1:
			return el->style->marginLeft + el->style->borderLeft + el->style->paddingLeft + el->style->width
				+ el->style->paddingRight + el->style->borderRight + el->style->marginRight;
		case 2:
			return el->style->marginTop + el->style->borderTop + el->style->paddingTop + el->style->height
				+ el->style->paddingBottom + el->style->borderBottom + el->style->marginBottom;
		case 3:
			return el->style->marginLeft + el->style->borderLeft + el->style->paddingLeft
				+ el->style->paddingRight + el->style->borderRight + el->style->marginRight;
		case 4:
			return el->style->marginTop + el->style->borderTop + el->style->paddingTop
				+ el->style->paddingBottom + el->style->borderBottom + el->style->marginBottom;
	}

	return 0;
}

//#############################################################################################################################
//
//                                             W i n d o w   O b j e c t
//
//#############################################################################################################################

void Window_init(void)
{
	window = (Window *)malloc(sizeof(Window));
	allocTracer_add(sizeof(Window),2);
	window->html = NULL;
	window->stylesheet = 0;

	window->width = 0;
	window->height = 0;
}

void Window_resize(uint width, uint height)
{
	if(width > 255) width = 255;
	if(height > 255) height = 255;

	window->width = width;
	window->height = height;

	Window_resolve();
}

void Window_setHTML(struct DomElement *p_html)
{
	if(!window->html) window->html = p_html;
}

const char *Window_workingDirectory(void)
{
	return window->directory;
}

void Window_setWorkingDirectory(const char *cwd)
{
	window->directory = cwd;
}

void Window_addStylesheet(reference_t s)
{
	if(window->stylesheet == 10) return;
	int length = s.end - s.begin;
	char *ptr = (char *)malloc(length+1);
	strncpy(ptr,s.begin,length);
	ptr[length] = 0;

	window->stylesheets[window->stylesheet++] = ptr;
	allocTracer_add(length+1,2);
}

/*
void Window_DEBUGWEB(void)
{
	if(window->html) DOMElement_DEBUGWEB(window->html,-1), Debug("\n"), DOMElement_DEBUGWEB(window->html,0);
}
*/

void Window_quit(void)
{
	Window_clear();
	free(window);
}

void Window_resolve(void)
{
	if(!window->html) return;

	window->html->style->top = 0;
	window->html->style->left = 0;
	window->html->style->maxWidth = window->width;

	DomElement_resolve(window->html);
}

void Window_clear(void)
{
	int i;

	for(i=0;i<window->stylesheet;i++)
		if(window->stylesheets[i]) free(window->stylesheets[i]);
	window->stylesheet = 0;

	if(window->html) DomElement_free(window->html);
	window->html = NULL;

	if(window->directory) free(window->directory);
}

//#############################################################################################################################
//
//                                      N o n - r e c u r s i v e   p a r s e r
//                                                Used by CSS system
//
//#############################################################################################################################

struct DomElement *Dom_parse(int initialize)
{
	static struct DomElement *ptr = NULL;
	int i;

	if(initialize)
	{
		ptr = NULL;
		return ptr;
	}
	if(!ptr)
	{
		ptr = window->html;
		return ptr;
	}

	if(ptr->childnum) { ptr = ptr->children[0]; return ptr; }
	else do
	{
		struct DomElement *old = ptr;
		i = 0;  
		ptr = ptr->parent;
		if(!ptr) return NULL;

		while(i<ptr->childnum && ptr->children[i]!=old) i++; i++;
	} while(i==ptr->childnum);

	ptr = ptr->children[i];
	return ptr;
}
