#ifndef _DOM_H
	#define _DOM_H

/*
	Header inclusions.
*/

#include "tex-0.9.h"
#include <stdref.h>



/*
	Type declarations.
*/

// Just lightening following declarations.
#ifndef _UINT
	#define _UINT
	typedef unsigned uint;
#endif // _UINT



/*
	Composed type definitions.
*/

/**
 *	DomStyle
 *
 *	This structure handles everything that has to do with styling.
 *	Comments represent the possible values for the property, in order of
 *	values. When there is no comment, then the corresponding property is
 *	associated a numeric value or pointer.
 *
 *	TODO
 *	-	Define DomStyle as an union so that simple tags do not carry
 *		useless style information.
 *	-	Use a 12-bit bit field for border parameters handling.
 */

struct DomStyle
{

	uint background		:1;	// white, black
	uint color		:1;	// white, black

	uint overflowX		:2;	// visible, hidden, scroll, auto
	uint overflowY		:2;	// visible, hidden, scroll, auto
	uint resize		:2;	// none, horizontal, vertical, both

	uint display		:2;	// none, inline, block, inline-block
	uint position		:2;	// static, absolute, fixed, relative
	uint cssFloat		:2;	// none, left, right
	uint verticalAlign	:2;	// top, middle, bottom

	uint top		:12;
	uint right		:12;
	uint bottom		:12;
	uint left		:12;

	uint width		:12;
	uint minWidth		:12;
	uint maxWidth		:12;
	uint height		:12;
	uint minHeight		:12;
	uint maxHeight		:12;

	uint paddingTop		:8;
	uint paddingRight	:8;
	uint paddingBottom	:8;
	uint paddingLeft	:8;

	uint borderTop		:8;
	uint borderRight	:8;
	uint borderBottom	:8;
	uint borderLeft		:8;
	uint borderRadius	:8;

	uint marginTop		:8;
	uint marginRight	:8;
	uint marginBottom	:8;
	uint marginLeft		:8;

	uint borderSpacingX	:8;
	uint borderSpacingY	:8;

	uint fontSize		:8;
	int  letterSpacing	:8;
	uint lineHeight		:8;
	int  textIndent		:8;

	uint cursor		:3;	// crosshair, default, help, move,
					// none, pointer, progress, text
	uint visibility		:1;	// hidden, visible
	uint listStyle		:4;	// 1  01  a  A  i  I  x  o  .  -  --

	uint textAlign		:2;	// left, center, right, justify
	uint textDecoration	:2;	// none, underline, overline,
					// line-trough
	uint whiteSpace		:2;	// normal (wrap), nowrap, pre, pre-wrap
	uint borderCollapse	:1;	// separate, collapse
	uint emptyCells		:1;	// show, hide

	uint tabSize		:4;
	uint wordSpacing	:4;

	/*
		Other elements.

	char *backgroundImage;
	*/
};

/**
 *	DomAttribute
 *
 *	This simple structure stores the content of a tag attribute, by name
 *	and value.
 */

struct DomAttribute
{
	char *name;
	char *value;
};

/**
 *	DomElement
 *
 *	This structure represents a node in the DOM element tree. This is the
 *	mainly manipulated structure in the interpreter.
 */

struct DomElement
{
	/*
		Basic element properties.
	*/

	// NUL-terminated tag name (dynamically allocated).
	char *tagName;
	// Style structure pointer (dynamically allocated).
	struct DomStyle *style;
	// Node data.
	union
	{
		// Text content, dynamically allocated.
		char *textNode;
		// Loaded Tex expression (the code is not saved).
		struct Tex_Expression *tex;
	};



	/*
		Children and attributes.
	*/

	// Children and attribute pointer arrays.
	struct DomElement **children;
	struct DomAttribute **attributes;

	// Children and attribute numbers counters.
	uint attrnum	:8;
	uint childnum	:8;



	/*
		Relations.
	*/

	// Parent node.
	struct DomElement *parent;
	// Previous sibling.
	struct DomElement *previous;



	/*
		Possible useful data for later.

	unsigned short accessKey;

	uint scrollWidth	:12;
	uint scrollHeight	:12;
	uint scrollTop		:12;
	uint scrollLeft		:12;
	*/
};

typedef struct
{
	struct DomElement *html;
	uint width		:8;
	uint height		:8;

	const char *directory;
	const char *stylesheets[10];
	int stylesheet		:8;
} Window;



/*
	Some useful information... (WH = Width/Height, TL = Top/Left)
	clientWH = inner + padding - scrollbars.
	offsetWH = inner + padding + border.
	offsetTL = coords + margin.
	scrollWH = overflow + padding - scrollbars.
	scrollTL = scrollbars offset.
*/



/*
	Function prototypes.
*/

// DomStyle manipulation functions.
struct DomStyle *DomStyle_new(void);

// DomAttribute manipulation functions.
struct DomAttribute *DomAttribute_new(reference_t name, reference_t value);
void DomAttribute_free(struct DomAttribute *attribute);

// DomElement manipulation functions.
struct DomElement *DomElement_new(reference_t name, reference_t args,
	struct DomElement *parent, struct DomElement *previous);
void DomElement_free(struct DomElement *element);
struct DomElement *DomElement_addChild(struct DomElement *element,
	struct DomElement *child);
struct DomElement *DomElement_getChild(struct DomElement *element, int index);
void DomElement_setTextNode(struct DomElement *element, reference_t text);
const char *DomElement_getAttribute(struct DomElement *element,
	const char *attribute_name);
void DomElement_resolve(struct DomElement *element);

int DomElement_get(struct DomElement *element, int value);

// Non-recursive tree parser.
struct DomElement *Dom_parse(int initialize);

	#define DOMElement_totalWidth(e)  DomElement_get(e,1)
	#define DOMElement_totalHeight(e) DomElement_get(e,2)
	#define DOMElement_tokenWidth(e)  DomElement_get(e,3)
	#define DOMElement_tokenHeight(e) DomElement_get(e,4)

#endif // _DOM_H
