#include <WebCalc.h>

#define isspace(c) ((c)==0x20 || ((c)>=0x09 && (c)<=0x0D))
#define forward(p); while(isspace(*(p))) p++;
#define backward(p); while(isspace(*(p))) p--;
#define backwards(p,s,c); while((p) != (s) && *(p) != (c)) p--;

/**
 *	display()
 *
 *	Vertically displays the document at the given offset. Offset is suppo-
 *	ded to be negative (it represents the row where the document is drawn).
 *	Also draws the scrollbar.
 *
 *	@param	offset	Display offset.
 */

void display(int offset)
{
	// Document height.
	int y = 0;
	// Using an incremental counter, and scrollbar parameters.
	int i, n, l;

	// Painting the document.
	Disp_paint(window->html,0,offset);

	// Getting the document's height.
	y = window->html->style->height;
	// Exiting if there is no scrollbar to draw.
	if(y <= 64) return;

	// Computing the scrollbar offset.
	n = (-offset*64)/y;
	// Computing the scrollbar length.
	l = 4096/y;
	// Drawing the scrollbar.
	for(i=n;i<=n+l;i++) dpixel(127,i,1);
}

void Disp_paint(struct DomElement *el, int x, int y)
{
	/*
	**  This function paints an element at absolute position (x;y). This absolute position can seem weird, but as long as
	**  the element positionning is made relatively to the parent, the absolute offset is inherited and it is possible to
	**  display the document at any screen offset.
	*/

	// Incremental counters, loop limits.
	uint i,j, beg,end;

	// Only drawing elements that are visible in the screen area.
	if(y > 63 || y + DOMElement_totalHeight(el) < 0) return;

	// Drawing top and bottom borders.
	if(el->style->borderTop || el->style->borderBottom)
	{
		beg = el->style->marginLeft;
		end = DOMElement_totalWidth(el) - el->style->marginRight -1;
		for(j=0;j<el->style->borderTop;j++) for(i=beg;i<=end;i++)
			dpixel(x+i,y+el->style->marginTop+j,1);
		for(j=0;j<el->style->borderBottom;j++) for(i=beg;i<=end;i++)
			dpixel(x+i,y+DOMElement_totalHeight(el)-el->style->marginBottom-j-1,1);
	}

	// Drawing left and right borders.
	if(el->style->borderLeft || el->style->borderRight)
	{
		beg = el->style->marginTop;
		end = DOMElement_totalHeight(el) - el->style->marginBottom -1;
		for(j=0;j<el->style->borderLeft;j++) for(i=beg;i<=end;i++)
			dpixel(x+el->style->marginLeft+j,y+i,1);
		for(j=0;j<el->style->borderRight;j++) for(i=beg;i<=end;i++)
			dpixel(x+DOMElement_totalWidth(el)-el->style->marginRight-j-1,y+i,1);
	}

	// Drawing background.
	if(el->style->background)
	{
	}

	// Update absolute coordinates depending on the margins, borders and padding of the element.
	x += el->style->marginLeft + el->style->borderLeft + el->style->paddingLeft;
	y += el->style->marginTop + el->style->borderTop + el->style->paddingTop;

	// Drawing a math formula if the tag name matches "math".
	if(!strcmp(el->tagName,"math")) tex_draw(el->tex,x,y);
	// Drawing text if the element is a text container.
	else if(el->textNode) Disp_text(el,2,x,y);
	// Painting the children.
	for(i=0;i<el->childnum;i++) Disp_paint(el->children[i],x+el->children[i]->style->left,y+el->children[i]->style->top);
}

void Disp_text(struct DomElement *el, int action, int x, int y)
{
	/*
	**  This function computes how the text will be arranged if it is drawn. It separates it in lines defined by the element
	**  max width set in el->style->maxWidth or the width set in el->style->width if the first is zero. If both are zero, the
	**  function returns immediatly.
	**
	**  Value for `action` :
	**  0  Doesn't do anything.
	**  1  Sets the element width and height.
	**  2  Draws the text.
	*/

	// Getting the text.
	char *text = el->textNode;
	// Pointers beg and end will define each word.
	const char *beg=text, *end=beg, *nextWord=beg;

	// Font to use.
	Font *font = font_nearest(el->style->fontSize);

	// Word spacing property.
	const int wordSpacing = (el->style->wordSpacing ? el->style->wordSpacing : 4);
	// Line height property.
	const int lineHeight = (el->style->lineHeight ? el->style->lineHeight :
		font->height) + 1;

	// These pointers will indicate where each line begins and ends.
	const char *lineBeg=text, *lineEnd=text;
	// Will contain the width of each evaluated word and line.
	int wordLength, lineLength=0;

	// Minimum width and height of the element.
	int width=0, height=0;

	if(y > 64 || y+el->style->height < 0) return;
	if(!(el->style->maxWidth | el->style->width)) return;

	/*
	// Debugging some function parameters.
	Debug("Text displaying\n");
	Debug("===============\n\n");
	Debug("Tag name    : %s\n",el->tagName);
	Debug("Coordinates : %d;%d\n",x,y);
	Debug("Font id.    : %d\n",el->style->font);
	Debug("Font size   : %d*%d\n", (3), (5));
//	Debug("Font size   : %d*%d\n",Txt_Width(el->style->font), Txt_Height(el->style->font));
	Debug("Spacing     : letter=%d, word=%d\n",el->style->letterSpacing,wordSpacing);
	Debug("Available   : %dpx\n",el->style->maxWidth);
	Debug("\n");
	*/

	// Main working loop -- each iteration matches a word.
	while(1)
	{
		// Trying to find a blank space.
		nextWord = strchr(beg,' ');
		// If a blank space exists.
		if(nextWord)
		{
			// Set end value at the end of the desired word.
			end = nextWord;
			// Move forward to next word.
			forward(nextWord);
		}
		// Otherwise, end is at the end of the string.
		else for(end=beg;*end;end++);

		// Calculate the length of the selected word.
		wordLength = font_length_pointer(beg,end,font);

		// Check if the word can be inserted at the end of the line.
		if(lineLength + wordSpacing + wordLength <= el->style->maxWidth
			- DOMElement_tokenWidth(el))
		{
			// Updating lineLength and lineEnd.
			if(lineLength != 0) lineLength += wordSpacing;
			lineLength += wordLength;
			lineEnd = end;
		}
		// If not, draw the line and reset the counters.
		else
		{
			// Update height (line separator).
			if(height) height++;

			// If desired, draw the text line.
			if(action==2) Disp_textDraw(lineBeg, lineEnd,
				el->style, x, y+height, lineLength);

			// Resetting the counters.
			lineBeg = beg;
			lineEnd = end;

			// Updating the minimum width and height and resetting
			// the line length.
			if(lineLength > width) width = lineLength;
			height += lineHeight;
			lineLength = wordLength;
		}

		// Exiting if the word was at the end of the string.
		if(!*end) break;
		// Updating beg using the nextWord pointer.
		beg = nextWord;
	}

	// Display the last line. 
	if(height) height++;
	if(action==2) Disp_textDraw(lineBeg,lineEnd,el->style,x,y+height,lineLength);
	height += lineHeight;

	// Set the width and height of the element.
	if(action==1)
	{
		el->style->width = width;
		el->style->height = height;
	}

	/*
	// Debugging some more.
	Debug("\n");
	Debug("Used        : %dpx\n",width);
	Debug("Height      : %dpx\n",height);
	*/
}

uint Disp_textLength(const char *text, const char *end, int fontWidth, int letterSpacing)
{
	/*
	**  This function returns the length of a word using the font width and the letter spacing. It does not add any whitespace,
	**  just computes how many pixels are strictly needed to draw the string.
	**  The string is represented as a beginning and an end pointer because it is used by Disp_text() that works this way (to
	**  prevent the program from modifying the original given string).
	**  Having a function to do this operation may be considered as useless, but it will allow to leave Disp_text() unchanged
	**  if non-monospaced fonts went to be implemented.
	*/

	// Parser pointer, word length.
	const char *parser = text;
	uint length = 0;

	// Main loop, by character.
	while(parser != end)
	{
		// Add the letter width.
		length += fontWidth;

		// Exit if at end of the string.
		if(++parser == end) break;
		// Add the spacing.
		length += letterSpacing;
	}

	// Return the computed length.
	return length;
}

void Disp_textDraw(const char *beg, const char *end, struct DomStyle *style,
	int x, int y, int lineLength)
{
	/*
	**  This function draws the given line of text (delimited by beg and end) at the coordinates given as arguments. As
	**  extra information, it gets the line length (just not to compute it two times) for text alignment and the element
	**  style structure, which gives it various informations.
	*/

	// Text parser, will move between beg and end.
//	const char *parser = beg;
	// Font to use.
	Font *font = font_nearest(style->fontSize);
	// Mine height and word spacing properties.
//	const uint lineHeight = style->lineHeight ? style->lineHeight : font->height;
//	const uint wordSpacing = style->wordSpacing ? style->wordSpacing : 4;

	// Setting x depending on alignment.
	if(style->textAlign == 1) x = x + ((style->width - lineLength) >> 1);
	if(style->textAlign == 2) x = x + style->width - lineLength;

//	if(style->textAlign==1) Debug("length %d, style->width %d, x %d\n",lineLength,style->width,x);

	font_text_pointer(x,y,beg,end,font);
}

/**********************************************************/

void disp_bitmap(int x, int y, const unsigned char *b, int w, int h)
{
	unsigned char v;
	int i,j;

	for(i=0;i<h;i++) for(j=0;j<w;j++)
	{
		v = (*(b+i*(((w-1)>>3)+1)+(j>>3))>>(~j&7))&1;
		dpixel(x+j,y+i,v);
	}
}
