/**
 *	CSS.c
 *	Parsing system
 */

#include <WebCalc.h>

#define isspace(c) ((c)==0x20 || ((c)>=0x09 && (c)<=0x0D))
#define forward(p); while(isspace(*(p))) p++;
#define backward(p); while(isspace(*(p))) p--;
#define backwards(p,s,c); while((p) != (s) && *(p) != (c)) p--;

static int atoi(reference_t ref)
{
	// Variables declaration.
	const char *str = ref.begin;
	int n=0, sign=1;

	// Parsing string until the first glyph.
	while((*str==' ' || *str=='\t' || *str=='\n') && str < ref.end) str++;
	if(str == ref.end) return -1;

	// Checking sign.
	if(*str=='-') sign=-1, str++;
	else if(*str=='+') sign=+1, str++;

	// Reading digits.
	while(*str >= '0' && *str <= '9' && str < ref.end)
	{
		n += *str - '0';
		if(++str < ref.end) n *= 10;
	}

	// Returning the final value.
	return n*sign;
}

/**
 *	css_parse_all()
 *
 *	Thought to be called after the dom tree has been loaded. Style is
 *	applied immediatly and without priority operations, which saves the
 *	memory that would have been used to store all the rules.
 *	There is no support for pseudo-selectors now, because the would need to
 *	store the style rules, and I have no efficient solution for that.
 *	The function loads and parses all linked stylesheets, one by one and in
 *	order. It only supports relative paths.
 */


void css_parse_all(void)
{
	// Using an iterator, a file descriptor and a file size.
	int i, fd, size;
	// Using a buffer for filename a pointer for loaded data.
	char file[40], *styledata;

	// Selecting stylesheets.
	for(i=0;i<window->stylesheet;i++)
	{
		// Creating the filename.
		strcpy(file,Window_workingDirectory());
		strcpy(file+strlen(file),window->stylesheets[i]);

		// Opening the file.
		fd = memory_openfile(file,_OPENMODE_READ);
		if(fd < 0) continue;

		// Allocating the memory.
		size = memory_filesize(fd);
		styledata = malloc(size+1);
		if(!styledata)
		{
			memory_closefile(fd);
			continue;
		}

		// Loading and parsing the file.
		memory_readfile(fd,styledata,size,-1);
		css_parse(styledata);

		// Freeing data and closing the file.
		free(styledata);
		memory_closefile(fd);
	}
}

/**
 *	css_parse()
 *
 *	This function is the main css sequential parser. It is designed to
 *	always succeed to parse and interpret a stylesheet, what implies that
 *	anything unrecognized will be ignored.
 *
 *	@param	data	Loaded stylesheet data, as ascii text.
 */

void css_parse(const char *data)
{
	// Using a temporary pointer;
	char *end;
	// Using a rule selector and references for properties and values.
	reference_t selector, property, value;

	// Every loop iteration should match a rule.
	while(1)
	{
		// Seeking selector.
		forward(data);
		if(!*data) break;

		// Selecting the selector, exiting if non-ended.
		end = strchr(data,'{');
		if(!end) break;
		end--;
		while(isspace(*end) && end!=data) end--;

		// Copying the selector.
		selector.begin = data;
		selector.end = end + 1;

		// Getting to the beginning of interesting data.
		data = strchr(data,'{') + 1;

		// Rule analysis.
		while(1)
		{
			// Isolating property.
			forward(data);
			end = strchr(data, ':') - 1;
			backward(end);

			// Copying property.
			property.begin = data;
			property.end = end + 1;

			// Isolating value.
			data = strchr(data, ':') + 1;
			forward(data);
			if(!*data || *data == '}') break;
			end = strchr(data, ';') - 1;
			backward(end);

			// Copying value.
			value.begin = data;
			value.end = end + 1;

			css_target(selector, property, value);
			data = strchr(data, ';') + 1;
			forward(data);

			// Breaking the loop when the rule ends.
			if(!*data || *data == '}') break;
		}

		// Exiting if the file has been totallt parsed.
		if(!*data) break;

		// At this point, *data is '}'.
		data++;
	}
}

/**
 *	css_target()
 *
 *	This function selects all the elements that matches the given selector,
 *	and calls css_apply() for all of these elements, in order to apply the
 *	rule.
 *
 *	@param	selector	Element selector, extracted from the sheet.
 *	@param	property	The property to set.
 *	@param	txtvalue	The value to be assigned to the property.
 */

void css_target(reference_t selector, reference_t property, reference_t value)
{
	// Using a pointer for non-recursive parsing on the dom tree.
	struct DomElement *element;
	// Using a reference to take selector's targets one by one.
	reference_t target;
	// Using an end pointer to reference data in the string.
	const char *ptr = selector.begin, *end = ptr;

	// Parsing targets, separated by a comma (',').
	while(ptr < selector.end)
	{
		end = ptr;
		while(end < selector.end && *end != ',') end++;
		target.begin = ptr;
		target.end = end;
		target = reftrim(target);

		// Targetting the elements.
		Dom_parse(1);
		while((element = Dom_parse(0)))
		{
			// Applies the property if the element is targetted.
			if(css_matches(element, target)) continue;
			css_apply(element, property, value);
		}

		// Preparing the next iteration.
		ptr = end + 1;
		forward(ptr);
		if(*ptr == ',') { ptr++; forward(ptr); }
	}
}

/**
 *	css_matches()
 *
 *	Tests if an element matches a CSS selector. Returns 0 if the element
 *	matches and 1 in any other case.
 *	For now, it only supports tag-name selectors (such as `h1` or `div a`)
 *	and immediate child selector `>` (which *must* be separated with
 *	spaces).
 *	Pseudo-selectors, classes and ids are detected but ignored.
 */

int css_matches(struct DomElement *element, reference_t target)
{
	// Using a parent pointer.
	struct DomElement *parent;
	// Using reference string parsers.
	const char *beg, *end;
	// Testing mode for the current tag (0:tag, 1:parent, 2:ancester).
	int mode = 0;
	// Using a temporary pointer and reference for parsing.
	reference_t ref;

	// Initializing the parsing pointers.
	end = target.end;

	// Isolating tags in parent/children relation.
	while(1)
	{
		// Looking for the beginning of the word end points to.
		beg = end - 1;
		while(!isspace(*beg) && beg > target.begin) beg--;
		if(beg != target.begin) beg++;

		// Referencing the word.
		ref.begin = beg;
		ref.end = end;

		// Testing match depending on the testing mode.
		switch(mode)
		{
			// Mode 0 : testing tag name.
			case 0:
				if(refscmp(ref, element->tagName)) return 1;
				mode = 2;
				break;

			// Mode 1 : tesing immediate parent tag name.
			case 1:
				if(refscmp(ref, element->parent->tagName))
					return 1;
				element = element->parent;
				mode = 2;
				break;

			// Mode 2 : testing a parent's tag name, and looking
			// for `>`.
			case 2:
				if(!refscmp(ref, ">")) { mode = 1; break; }
				parent = element;
				while((parent = parent->parent))
				{
					if(!refscmp(ref, parent->tagName))
						break;
				}
				if(parent) { element = parent; break; }
				return 1;
		}

		// Exiting if all parents have been verified.
		if(beg == target.begin) break;

		// Moving end point to the previous word.
		end = beg-1;
	}

	// If loops successfully ends, the element matches.
	return 0;
}

/**
 *	css_apply()
 *
 *	Applies the value for the given property on the given element. It
 *	mainly makes the link between property names and offsets in the
 *	structure. It then sets the correct integer value after testing the
 *	strings (in the case the values have a literal representation). 
 */

void css_apply(struct DomElement *element, reference_t p, reference_t v)
{
	// Using an unsigned value and an int to store atoi()'s results.
	const char *property = p.begin, *value = v.begin;
	unsigned uv;
	int x;

	// Handling background and color properties.
	if(!refscmp(p, "background") || !refscmp(p, "color"))
	{
		if(!refscmp(v, "white")) uv = 0;
		else if(!refscmp(v, "black")) uv = 1;
		else return;
		// Checking which property was changed.
		if(*property == 'b') element->style->background = uv;
		else element->style->color = uv;
	}

	// Handling overflow in two dimensions.
	else if(!refscmp(p, "overflow-x") || !refscmp(p, "overflow-y"))
	{
		x = css_index(v, 4, "visible", "hidden", "scroll", "auto");
		if(x == -1) return;
		if(property[9] == 'x') element->style->overflowX = x;
		else element->style->overflowY = x;
	}

	// Handling resizing options.
	else if(!refscmp(p, "resize"))
	{
		x = css_index(v, 4, "none", "horizontal", "vertical",
			"both");
		if(x+1) element->style->resize = x;
	}

	// Handling display mode.
	else if(!refscmp(p, "display"))
	{
		x = css_index(v, 4, "none", "inline", "block",
			"inline-block");
		if(x+1) element->style->display = x;
	}

	// Handling element positionning.
	else if(!refscmp(p, "position"))
	{
		x = css_index(v, 4, "static", "absolute", "fixed",
			"relative");
		if(x+1) element->style->position = x;
	}

	// Handling element floating mode. Erases positionning.
	else if(!refscmp(p, "float"))
	{
		x = css_index(v, 3, "none", "left", "right");
		if(x+1) element->style->cssFloat = x;
	}

	// Handling element's vertical align.
	else if(!refscmp(p, "vertical-align"))
	{
		x = css_index(v, 3, "top", "middle", "bottom");
		if(x+1) element->style->verticalAlign = x;
	}

	// Handling integer values (1) : position and size.
	else if(!(refscmp(p, "top") && refscmp(p, "right")
		&& refscmp(p, "bottom") && refscmp(p, "left")
		&& refscmp(p, "width") && refscmp(p, "height")))
	{
		int x = atoi(v);
		switch(*property)
		{
			case 't': element->style->top = x; break;
			case 'r': element->style->right = x; break;
			case 'b': element->style->bottom = x; break;
			case 'l': element->style->left = x; break;
			case 'w': element->style->width = x; break;
			case 'h': element->style->height = x; break;
		}
	}

	// Handling integer values (2) : size limits.
	else if(!(refsncmp(p, "min", 3) && refsncmp(p, "max", 3))
		&& !(strcmp(property + 3, "-width")
		&& strcmp(property + 3, "-height")))
	{
		uint x = atoi(v);
		int d = ((*(property+1)=='i')<<1) | (*(property+4)=='w');

		switch(d)
		{
			case 0: element->style->maxHeight = x; break;
			case 1: element->style->maxWidth = x; break;
			case 2: element->style->minHeight = x; break;
			case 3: element->style->minWidth = x; break;

		}
	}

	// Handling integer values (3) : text formatting.
	else if(!refscmp(p, "letter-spacing"))
	{
		uint x = atoi(v);
		element->style->letterSpacing = x;
	}

	else if(!refscmp(p, "line-height"))
	{
		uint x = atoi(v);
		element->style->lineHeight = x;
	}

	// Handling font size.
	else if(!refscmp(p, "font-size"))
	{
		uint x = atoi(v);
		element->style->fontSize = x;
	}

	// Handling text alignment.
	else if(!refscmp(p, "text-align"))
	{
		x = css_index(v, 4, "left", "center", "right", "justify");
		if(x+1) element->style->textAlign = x;
	}

	// Handling table border spaces.
	else if(!refscmp(p, "border-spacing"))
	{
		reference_t ref;
		uint x,y;

		char *tmp = strchr(value, ' ');

		if(tmp)
		{
			ref.begin = p.begin;
			ref.end = tmp;
			x = atoi(ref);

			forward(tmp);
			ref.begin = tmp;
			ref.end = p.end;
			y = atoi(ref);
		}
		else
		{
			x = atoi(p);
			y = x;
		}

		element->style->borderSpacingX = x;
		element->style->borderSpacingY = y;
	}

	// Handling "everything else", which is padding, borders and margins.
	else
	{
		// This variable will determine which property is changed.
		int d = 0;
		uint x;

		// Checking property's family.
		if(!refsncmp(p, "padding", 7)) p.begin += 7;
		else if(!refsncmp(p, "border", 6)) d = 4, p.begin += 6;
		else if(!refsncmp(p, "margin", 6)) d = 8, p.begin += 6;
		else return;

		// Checking property's direction.
		if(!refscmp(p, "-right")) d |= 1;
		else if(!refscmp(p, "-bottom")) d |= 2;
		else if(!refscmp(p, "-left")) d |= 3;
		else if(refscmp(p, "-top")) return;

		// Getting value.
		x = atoi(v);

		// Applying modification.
		switch(d)
		{
			case 0: element->style->paddingTop = x; break;
			case 1: element->style->paddingRight = x; break;
			case 2: element->style->paddingBottom = x; break;
			case 3: element->style->paddingLeft = x; break;
			case 4: element->style->borderTop = x; break;
			case 5: element->style->borderRight = x; break;
			case 6: element->style->borderBottom = x; break;
			case 7: element->style->borderLeft = x; break;
			case 8: element->style->marginTop = x; break;
			case 9: element->style->marginRight = x; break;
			case 10: element->style->marginBottom = x; break;
			case 11: element->style->marginLeft = x; break;
		}
	}
}

/**
 *	css_index()
 *
 *	Returns the index of the given string in the string list given as a
 *	variable argument list. Returns -1 if not found.
 */

int css_index(reference_t value, int n, ...)
{
	// Using a va_list object and an incremental counter.
	va_list options;
	int i;

	// Initializing the va_list.
	va_start(options,n);

	// Looping through the arguments.
	for(i=0; i<n; i++)
		if(!refscmp(value, va_arg(options, const char *))) break;

	// If there is no match, returning -1.
	va_end(options);
	return (i==n ? -1 : i);
}
