#include <WebCalc.h>

int isvalid(int c)
{
	return (c>='0' && c<='9') || (c>='A' && c<='Z') || (c>='a' && c<='z')
		|| c=='-';
}

#define isspecs(c) ((c)=='/' || (c)=='!')
#define isspace(c) ((c)==' ' || (c)=='\t' || (c)=='\n')

// Block type enumeration definition.
enum block_t
{
	block_tag  = 0,
	block_text = 1
};

// Element structure definition.
struct el_t
{
	// Opening tag name reference (closing tag carries no info).
	reference_t name;
	// Tag arguments.
	reference_t args;
	// Text content, if needed.
	reference_t text;
	// Associated element.
	struct DomElement *dom;
	// Linked stack pointer.
	struct el_t *down;
};

static struct el_t *top = NULL;


/**
 *	html_text()
 *
 *	Indicates if the given tag name is associated with a text content, or
 *	not.
 *
 *	@param	name	Reference on tag name.
 *
 *	@return		1 if the tag is supposed to have a text content, 0
 *			otherwise.
 */

static int html_text(reference_t name)
{
	const char *texts[] = { "p", "h1", "h2", "h3", "h4", "h5", "h6",
		"math", "center", "right", NULL };
	const char **ptr = texts;

	while(*ptr) if(!refscmp(name,*ptr++)) return 1;
	return 0;
}

static int html_conf(reference_t name)
{
	const char *confs[] = { "head", "link", NULL };
	const char **ptr = confs;

	while(*ptr) if(!refscmp(name,*ptr++)) return 1;
	return 0;
}

static void html_configure(reference_t name, reference_t args)
{
	html_args(&args, NULL);
	reference_t arg, value;

	if(!refscmp(name,"link"))
	{
		reference_t rel = { 0,0 }, href = { 0,0 };

		while(1)
		{
			html_args(&arg, &value);
			if(refnull(arg)) break;
			else if(!refscmp(arg, "rel")) rel = value;
			else if(!refscmp(arg, "href")) href = value;
		}

		if(refnull(rel) || refnull(href)) return;
		if(!refscmp(rel, "stylesheet")) Window_addStylesheet(href);
	}
}

static void html_create(struct el_t *el)
{
//	const char *ptr;

/* Log is disabled.
	// Debugging the tag parameters.
	fputs("{\n\t", output);
	ptr = el->name.begin;
	while(ptr < el->name.end) fputc(*ptr++, output);
	if(!refnull(el->args))
	{
		fputs("\n\t", output);
		ptr = el->args.begin;
		while(ptr < el->args.end) fputc(*ptr++, output);
	}
	if(!refnull(el->text))
	{
		fputs("\n\t\"", output);
		ptr = el->text.begin;
		while(ptr < el->text.end) fputc(*ptr++, output);
		fputc('"', output);
	}
	fputs("\n}\n\n", output);
*/

	struct el_t *parent = el->down;

	if(parent)
	{
		el->dom = DomElement_new(el->name, el->args, parent->dom,
			DomElement_getChild(parent->dom, -1));
		DomElement_addChild(parent->dom, el->dom);
	}
	else el->dom = DomElement_new(el->name, el->args, NULL, NULL);

	if(!refnull(el->text)) DomElement_setTextNode(el->dom, el->text);

	if(!refscmp(el->name, "html")) Window_setHTML(el->dom);
}

/**
 *	html_event()
 *
 *	Treats an event that concerns the given block with the given type. An
 *	event may lead to one of the following operations :
 *	- adding an element to the stack
 *	- removing an element from the stack, creating the dom object
 *	- setting parameters of the top object
 *
 *	@param	block	Reference on the event block.
 *	@param	type	Type of the given block.
 *
 *	@return		Expected type for the next block.
 */

static enum block_t html_event(reference_t block, enum block_t type)
{
	// Using a temporary name and arguments references for validity tests.
	reference_t name, args;
	// Using a common pointer and a common integer.
	const char *ptr;
	int x;

	// Receiving an event for text content.
	if(type == block_text)
	{
		// Setting the content reference.
		if(top) top->text = block;
		// Returning block_tag as nature of next expected element.
		return block_tag;
	}



	// Only tag events reach this point.
	// Extracting the tag name.
	while(!isvalid(*block.begin) && !isspecs(*block.begin) &&
		block.begin < block.end) block.begin++;
	// Handling empty tags.
	if(block.begin == block.end) return block_tag;
	// Getting the name end.
	ptr = block.begin + isspecs(*block.begin);
	while(isvalid(*ptr)) ptr++;
	// Setting name reference.
	name.begin = block.begin;
	name.end = ptr;
	// Initializing tag validity checking operation.
	x = 0;
	ptr = name.begin + isspecs(*name.begin);
	// Checking tag validity. (e.g. '<//tag>', '<! >'). x is validity.
	while(ptr < name.end)
	{
		// Warning ! isvalid() is a macro, using *ptr++ would repeat
		// pointer incrementation !
		if(isvalid(*ptr)) x = 1;
		// Incrementing pointer.
		ptr++;
	}
	// A tag name is valid if it contains at least one valid character.
	if(!x) return block_tag;

	// Looking for a parameter string.
	args.begin = name.end;
	args.end = block.end - 1;
	args = reftrim(args);

	// Handling interpreter directives.
	if(*name.begin == '!')
	{
/* Log is disabled.
		// Outputting a string.
		fputc('#', output);
		ptr = name.begin + 1;
		while(ptr < name.end) fputc(*ptr++, output);
		fputs(" ", output);
		ptr = args.begin;
		while(ptr < args.end) fputc(*ptr++, output);
		fputs("\n\n", output);
*/
		// Returning a proper data type.
		return block_tag;
	}

	if(html_conf(name))
	{
		html_configure(name, args);
		return block_tag;
	}

	// Handling closing tags.
	if(*name.begin == '/')
	{
		name.begin++;
		if(html_conf(name)) return block_tag;

		struct el_t *el = top->down;
		// Non-textual tags are created when opened.
		if(!top->dom) html_create(top);
		free(top);
		top = el;

		return block_tag;
	}


	// Only opening tag events reach this point.
	// Allocating an object.
	struct el_t *el = (struct el_t *)malloc(sizeof(struct el_t));
	el->down = top;
	top = el;

	// Setting element parameters.
	el->name = name;
	el->args = args;
	el->text.begin = el->text.end = NULL;

	// Creating non-textual tags to ensure stack order.
	if(!html_text(name)) html_create(el);
	else el->dom = NULL;

	// Returning the correct expected data type.
	return html_text(name) ? block_text : block_tag;
}

/**
 *	html_next()
 *
 *	Returns a reference on the next element of the given type in the given
 *	code, defined by a begin and an end pointer.
 *	An element may be empty, particularly when a text element (such as <p>)
 *	does not contain any text.
 *
 *	@param	begin	Pointer on code.
 *	@param	limit	Pointer on the first byte that is outside the code.
 *	@param	type	Type of expected element.
 *
 *	@return		Reference on the element.
 */

static reference_t html_next(const char *begin, const char *limit,
	enum block_t type)
{
	// Using a reference.
	reference_t ref = { begin, begin };

	// Getting the next tag opening element.
	while(ref.end < limit && *ref.end != '<') ref.end++;
	// Returning if the given block type has the calculated limits.
	if(type == block_text) return ref;

	// Otherwise, skipping the first text prevents returning empty blocks.
	while(ref.end < limit && *ref.end != '>') ref.end++;
	// Increasing ref.end to comply with program conventions on references.
	if(ref.end != limit) ref.end++;
	// Returning the computed reference.
	return ref;
}

/**
 *	html_parse()
 *
 *	Parses the given html code, defined by a beginning and an end pointer.
 *	The end pointer must point to the first byte outside the code. In other
 *	words, it is supposed to be equal to `data + data_length`.
 *	The generated DOM is stored in an element tree whose entry point is
 *	global. All data is copied so the loaded code can be freed afterwards.
 *
 *	@param	data	Pointer on code data.
 *	@param	limit	Pointer on first byte outside the code data.
 */

void html_parse(const char *data, const char *limit)
{
	// Parsing pointer.
	const char *ptr = data;
	// Element reference.
	reference_t ref;
	// Block type.
	enum block_t type = block_tag;

	// Looping on elements until limit is reached.
	while(ptr < limit)
	{
		// Getting a reference on the next element.
		ref = html_next(ptr, limit, type);

		// Updating main pointer.
		ptr = ref.end;

		// Removing leading and ending spaces for tags.
		if(type == block_tag) ref = reftrim(ref);

		// Generating an event.
		type = html_event(ref, type);
	}
}

/**
 *	html_args()
 *
 *	Argument string interpreter. Meant to be used in several calls. A first
 *	call as html_args(&ref,NULL) initializes the operation, and every
 *	subsequent calls with two non-null pointers extracts an argument.
 *
 *	@param	name	Reference pointer. Arguments at initialization, address
 *			of argument name when extracting.
 *	@param	value	Reference pointer. NULL at initialization, address of
 *			argument value when extracting.
 */

void html_args(reference_t *name, reference_t *value)
{
	// Static reference for multiple calls.
	static reference_t args;
	// Parsing pointer.
	const char *ptr = args.begin;

	// Initialization.
	if(name && !value)
	{
		// Actually, name is meant to be the full argument string.
		args = *name;
		// Immediatly returning.
		return;
	}

	// Setting name. This code automatically handles empty names, etc.
	while(!isvalid(*ptr) && ptr < args.end) ptr++;
	name->begin = ptr;
	while(isvalid(*ptr) && ptr < args.end) ptr++;
	name->end = ptr;

	// Skipping spaces before '=' symbol, if existing.
	while(isspace(*ptr) && ptr < args.end) ptr++;
	if(ptr == args.end || *ptr != '=')
	{
		value->begin = value->end = ptr;
		return;
	}

	// Skipping '='.
	ptr++;
	// Skipping spaces after '='.
	while(isspace(*ptr) && ptr < args.end) ptr++;
	if(ptr == args.end)
	{
		value->begin = value->end = ptr;
		return;
	}

	// Getting the value.
	if(*ptr == '"')
	{
		value->begin = ++ptr;
		while(*ptr != '"' && ptr < args.end) ptr++;
		value->end = ptr;
	}
	else
	{
		while(!isvalid(*ptr) && ptr < args.end) ptr++;
		value->begin = ptr;
		while(isvalid(*ptr) && ptr < args.end) ptr++;
		value->end = ptr;
	}

	args.begin = ptr;
}

/**
 *	html_clear()
 *
 *	Clears the html module, that is, re-initializing all static pointers
 *	and previously kept data to ensure interpretation safety.
 */

void html_clear(void)
{
	// Using a stack element structure, and a temporary one.
	struct el_t *el = top, *tmp;

	// Clearing the stack.
	while(el)
	{
		tmp = el->down;
		free(el);
		el = tmp;
	}
	// Setting the top pointer.
	top = NULL;
}
