/**
 *	libfont version 3.1
 *
 *	Written by lephe.
 *	The library consists in this source file and its associated header
 *	file.
 *
 *	The library is released in public domain (CC0).
 *	See [http://creativecommons.org/about/cc0].
 */

// Header inclusion.
#include "libfont-3.1.h"

// Static symbols definition.
static struct Font *fonts[FONT_NUMBER] = { 0 };
static int font_number = 0;
static void (*dpixel)(int x, int y, int color) = NULL;

/**
 *	font_set_pixel_function()
 *
 *	Sets the function that shall be used to draw pixels. The first two
 *	arguments are the coordinates and the third is the color. Having color
 *	to 0 means white, any other value means black.
 *
 *	@param	func	Pointer to the pixel function.
 */

void font_set_pixel_function(void (*func)(int x, int y, int color))
{
	dpixel = func;
}

/**
 *	font_load()
 *
 *	Loads the given font in the font table. The argument is the address of
 *	loaded file data. Indeed, this version of the library is meant to be
 *	able to load fonts that are embedded in the executable file. Therefore,
 *	it only computes pointers, and does not copy data. So, loaded file data
 *	MUST NOT be freed after the font has been loaded. The font file format
 *	is made to store only useful information, so no memory is lost.
 *	This function returns a font structure pointer that shall be used with
 *	any function that requires to specify a font.
 *
 *	@param		address		objcopy symbol or address of loaded
 *					data.
 *	@param		error		Pointer to an error variable that is
 *					set (if not NULL) if an error occurs.
 *					If no error occurs, it is set to 0.
 *
 *	@return		Structure pointer or NULL if an error occured.
 *
 *	@warn		It is possible that structure members `italic` and
 *			`weight`, as well as the two bytes of `glypNumber` are
 *			swapped, depending on how the compiler handles the
 *			structure.
 *
 *	@doc		Error codes :
 *			-1	Maximum font number reached.
 *			-2	Alloc failure.
 */

struct Font *font_load(void *address, int *error)
{
	// Checking remaining space in the font array.
	if(font_number == FONT_NUMBER) { if(error) *error = -1; return NULL; }
	// Getting a better pointer.
	const char *data = (const char *)address;
	// Using an incremental counter and a dummy variable.
	int i, x;

	// Allocating the font structure and checking malloc failure.
	struct Font *f = malloc(sizeof(struct Font));
	if(!f) { if(error) *error = -2; return NULL; }
	// Adding the font to the global array.
	fonts[font_number++] = f;

	// Copying the meta data. This is necessary because if f pointed to
	// data, the double pointer at the end of the structure would override
	// existing font data.
	memcpy((void *)f,(const void *)data,12);
	// Swapping italic and weight members.
	x = (f->italic << 7) | f->weight;
	x = (x << 1) | (x & 128);
	f->italic = (x & 128) >> 7;
	f->weight = x & 127;
	// Swapping the lower and upper bytes of glyphNumber member.
	x = f->glyphNumber;
	x = ((x & 0xff00) >> 8) | ((x & 0x00ff) << 8);
	f->glyphNumber = x;

	// Allocating the glyph array. Properly exiting on alloc failure.
	f->glyphs = calloc(f->glyphNumber,sizeof(struct FontGlyph *));
	if(!f->glyphs) { if(error) *error = -2; free(f); return NULL; }

	// Advancing data pointer to glyph information.
	data += 12;

	// Reading glyphs.
	for(i=0;i<f->glyphNumber;i++)
	{
		// The last member, data, automatically points to the glyph
		// data.
		struct FontGlyph *g = (struct FontGlyph *)data;
		// Setting the glyph in the glyph array.
		f->glyphs[i] = g;

		// Computing the data length for the current glyph.
		x = (g->width * g->height + 7) >> 3;
		// Advancing data of the right offset.
		data += x + 4;
	}

	// Setting the error variable.
	if(error) *error = 0;
	// Returning the loaded font.
	return f;
}

/**
 *	font_free()
 *
 *	Frees the font given as argument. The glyph data is not unloaded since
 *	it may be part of the .data sections in the executable file.
 *	This functions only frees data allocated by font_init(), that is, glyph
 *	array and font structure. The original data must be freed by the user
 *	program if not part of the executable file.
 *	The freed font is also removed from the font database and may not be
 *	returned by font_nearest() anymore.
 *
 *	@param	font	Font to free.
 */

void font_free(struct Font *font)
{
	// Using an incremental counter.
	register int i = 0;

	// Ignoring the fonts before the interesting one.
	while(fonts[i] != font) i++;
	// Moving up the bottom of the array.
	while(i < font_number-1) fonts[i] = fonts[i+1], i++;
	// Removing the last array pointer (ensuring zero in all non-used
	// slots).
	fonts[font_number-1] = 0;
	// Updating the font number.
	font_number--;

	// Freeing the glyph array.
	free(font->glyphs);
	// Freeing the font pointer.
	free(font);
}

/**
 *	font_free_all()
 *
 *	Clears the font database and frees all the loaded fonts. No font may be
 *	returned by font_nearest() after this operation.
 *	Intended to be used when exiting the program. Do not unload fonts until
 *	you're done with them.
 */

void font_free_all(void)
{
	// Using a for loop would result in errors since font_number is reduced
	// by font_free().
	while(font_number) font_free(fonts[0]);
}

/**
 *	font_text_pointer()
 *
 *	Renders the text between pointer str and end at the given position
 *	(x;y), using the given font.
 *	This function basically allows to print a part of a string, for example
 *	a part of a loaded code, or user-input string, that is being
 *	interpreted. str must be lower than end. The character at adress end is
 *	not rendered.
 *	The glyphs are stored as a bit stream, that is, as if the glyph columns
 *	had been put together to make a big one that would have been
 *	represented pixel by pixel in binary.
 *
 *	@param	x	Left coordinate of the rendering rectangle.
 *	@param	y	Top coordinate of the rendering rectangle.
 *	@param	str	Top address of text to render.
 *	@param	end	End address of the text.
 *	@param	font	Font to use.
 */

void font_text_pointer(int x, int y, const char *str, const char *end,
	struct Font *font)
{
	// Using a font glyph pointer and a glyph data parser.
	struct FontGlyph *g;
	unsigned char v;
	// Using a character variable and two incremental counters.
	int c, i, j;
	// Using two variables to handle offset in bytes and bits.
	int oB, ob;

	// Returning if no pixel function was set.
	if(!dpixel) return;

	// Returning if the given information is not valid.
	if(!(font && str && end && end > str)) return;

	// Looping through the given string.
	while(str != end)
	{
		// Extracting the next character.
		c = *str++;

		// Looking for the corresponding glyph.
		for(i=0;i<font->glyphNumber;i++)
			if(font->glyphs[i]->character == c) break;
		// If no one is found, continue to the next iteration.
		if(i == font->glyphNumber) continue;
		// Saving the glyph address in a variable.
		g = font->glyphs[i];

		// Initialising the offsets. oB is 1 because the first byte is
		// read immediatly.
		oB = 1, ob = 0, v = g->data[0];

		// Drawing the pixels of the glyph.
		for(i=0;i<g->height;i++) for(j=0;j<g->width;j++)
		{
			// Setting the pixel at the right color.
			(*dpixel)(x+j,y+i,v & 128);
			// Getting the next bit.
			v <<= 1, ob++;
			// Switching to the next byte if the current one has
			// been totally read.
			if(ob == 8) v = g->data[oB++], ob = 0;
		}

		// Updating x-coordinate.
		x += g->width + 1;
	}
}

/**
 *	font_text()
 *
 *	Draws the given string at position (x;y) on the screen, using the given
 *	font. This function calls font_text_pointer() and just allows to print
 *	text the common way.
 *
 *	@param	x	Left coordinate of the rendering rectangle.
 *	@param	y	Top coordinate of the rendering rectangle.
 *	@param	str	NUL-terminated string to render.
 *	@param	font	Font to use.
 */

void font_text(int x, int y, const char *str, struct Font *font)
{
	// Using an incremental counter.
	register int i = 0;
	// Getting a pointer on the end of the text.
	while(str[i]) i++;
	// The NUL is ignored because it isn't printable.
	font_text_pointer(x,y,str,str+i,font);
}

/**
 *	font_length_pointer()
 *
 *	Computes the width that would span the given text if rendered with the
 *	given font. The text is represented by its top and bottom addresses to
 *	allow calculations on string parts. end must be greater that str.
 *
 *	@param	str	Top address of text.
 *	@param	end	End pointer.
 *	@param	font	Font to use.
 *
 *	@return		The width of the string, if rendered with the given
 *			font, or -1 if the given information is not valid.
 */

int font_length_pointer(const char *str, const char *end, struct Font *font)
{
	// Using a length variable.
	unsigned length = 0;
	// Using a character extractor and an incremental counter.
	int c, i;

	// Returning if the given information is not valid.
	if(!(font && str && end)) return -1;
	// Returning 0 if the text range is null.
	if(str == end) return 0;

	// Looping through the characters.
	while(str != end)
	{
		// Getting the next character.
		c = *str++;

		// Looking for an associated glyph.
		for(i=0;i<font->glyphNumber;i++)
			if(font->glyphs[i]->character == c) break;
		// If no one is found, continue to the next iteration.
		if(i == font->glyphNumber) continue;

		// Updating length variable using the glyph width. Adding
		// spacing.
		length += font->glyphs[i]->width + 1;
	}

	// Removing the last useless space and returning the computed length.
	return length - 1;
}

/**
 *	font_length()
 *
 *	Computes the width that is necessary to render the given text with the
 *	given font. This function uses a standard string and calls
 *	font_length_pointer() with a pointer to the first NUL character.
 *
 *	@param	str	The string whose length will be calculated.
 *	@param	font	The font to use.
 *
 *	@return		The computed width.
 */

int font_length(const char *str, struct Font *font)
{
	// Using an incremental counter.
	register int i = 0;
	// Getting a pointer on the first NUL character.
	while(str[i]) i++;
	// Delegating the calculation to font_length_pointer().
	return font_length_pointer(str,str+i,font);
}

/**
 *	font_length_character()
 *
 *	Extracts and returns the length of a single character in a given font.
 *	This function is not called by any other length-computing function.
 *
 *	@param	c	Character whose length is wanted.
 *	@param	font	Font to use.
 *
 *	@return		Width of the given character if it exists, 0 otherwise.
 */

int font_length_character(int c, struct Font *font)
{
	// Using an iterator.
	int i;

	// Looking for an associated glyph.
	for(i=0; i < font->glyphNumber; i++)
	{
		// If found, returning its width.
		if(font->glyphs[i]->character == c)
			return font->glyphs[i]->width;
	}

	// If no glyph exists for the given character, return 0.
	return 0;
}

/**
 *	font_nearest()
 *
 *	Returns a pointer on the loaded font, whose height is the nearest to
 *	the given argument. The retured font may be smaller, but never larger.
 *
 *	@param	size	The desired font size.
 *
 *	@return		The font in the database, whose height is nearest to
 *			the size argument, or NULL if no font has been loaded.
 */

struct Font *font_nearest(int size)
{
	// Getting a first pointer.
	struct Font *f = NULL;
	// Using an incremental counter and a height variable.
	int i, h;

	// Looping through the font array.
	for(i=0;i<font_number;i++)
	{
		// Getting the height of the next font.
		h = fonts[i]->height;
		// If the height is closer to the desired one, update the
		// pointer.
		if(!f || (f && h <= size && h > f->height)) f = fonts[i];
	}

	// Returning the font pointer.
	return f;
}
