#include <stdref.h>

#define isspace(c) ((c)==0x20 || ((c)>=0x09 && (c)<=0x0D))

/**
 *	refnull()
 *
 *	Tests if a reference points to a valid memory area.
 *
 *	@param	ref	Reference to test.
 *
 *	@return		1 if the reference is invalid or null, 0 otherwise.
 */

int refnull(reference_t ref)
{
	// Directly returning 1 if the reference is null, 0 otherwise.
	return (ref.begin >= ref.end);
}



/**
 *	refxtrct()
 *
 *	Extracts the data pointed by a reference and returns it as a
 *	dynamically-allocated text pointer. The text is NUL-ended.
 */

char *refxtrct(reference_t ref)
{
	// Storing the reference length in a variable.
	const int length = ref.end - ref.begin;
	// Using a text pointer.
	char *text;

	// Allocating memory.
	text = (char *)malloc(length + 1);
	// Handling alloc failure.
	if(!text) return NULL;

	// Copying the data.
	memcpy(text, ref.begin, length);
	// Setting the NUL ending character.
	text[length] = 0;

	// Returning the text.
	return text;
}



/**
 *	reftrim()
 *
 *	Reduces the given reference by excluding the beginning and ending
 *	spacing.
 *
 *	@param	ref	Original reference.
 *
 *	@return		Trimmed reference.
 */

reference_t reftrim(reference_t ref)
{
	// Reducing the end address until a non-space charater is reached.
	while(ref.end > ref.begin && isspace(*(ref.end-1))) ref.end--;
	// If the reference has become empty, returning it.
	if(ref.end == ref.begin) return ref;

	// Now that we ensured the reference contains non-space characters,
	// just going forward from the beginning.
	while(isspace(*ref.begin)) ref.begin++;
	// Returning the obtained reference.
	return ref;
}



/**
 *	refscmp()
 *
 *	Compares a reference with a string. Returns the result of strncmp()
 *	comparison, and -1 whenever the arguments are not valid or the lengths
 *	don't match.
 *
 *	@param	ref	Reference to compare to.
 *	@param	str	String to compare with.
 *
 *	@return		Result of strncmp() comparison, or -1 on error.
 */

int refscmp(reference_t ref, const char *str)
{
	// Using a variable to store the comparing length.
	int length;

	// Testing arguments validity.
	if(refnull(ref) || !str) return -1;

	// Getting the wanted length.
	length = strlen(str);
	// Returning 0 if the string length does not match the reference's.
	if(ref.end - ref.begin != length) return -1;

	// Returning the result of strncmp() comparison.
	return strncmp(ref.begin, str, length);
}



/**
 *	refs,cmp()
 *
 *	Compares a reference with a string in the limit of `length` characters.
 *
 *	@param	ref	Reference to compare to.
 *	@param	str	String to compare with.
 *	@param	length	Maximum number of characters to compare.
 *
 *	@return		Result of strncmp() comparison, or -1 on error.
 */

int refsncmp(reference_t ref, const char *str, int length)
{
	// Testing arguments validity.
	if(refnull(ref) || !str || length <= 0) return -1;

	// Returning the result of strncmp() comparison.
	return strncmp(ref.begin, str, length);
}
