/*
	Headers inclusion.
*/

#include "tex-0.9.h"



/*
	Macro functions.
*/

// Returns the greater of two values.
#define max(a, b) ((a) > (b) ? (a) : (b))
// Tests if a given character is a function opening character.
#define isfunc(c) ((c)=='\\' || (c)=='^' || (c)=='_')
// Tests if a given character is a bracket function identifier.
#define isbracket(c) ((c)=='(' || (c)==')' || (c)=='[' || (c)==']' \
	|| (c)=='{' || (c)=='}')
// Tests if a given character is alphanumeric. Independant of ctype for now.
#ifndef isalpha
	#define isalpha(c) (((c)>='a' && (c)<='z') || ((c)>='A' && (c)<='Z') \
	|| ((c)>='0' && (c)<='9'))
#endif // isalpha



/*
	Static variables and constants.
*/

// Pixel drawing function pointer.
static void (*dpixel)(int x, int y, int color) = NULL;
// Fonts for drawing common text and symbol (defined format).
static struct Font *fonts_text = NULL;
static struct Font *fonts_sybl = NULL;

// Static constant function and symbol names.
static const char *TEX_FUNCTIONS[] = {
	// Standard function.
	"frac", "sqrt", "vec", "sum", "prod", "int", "lim", "sup", "sub",
	// Brackets.
	"(", ")", "[", "]", "{", "}",
	// NULL terminator.
	NULL };
static const char *TEX_SYMBOLS[] = {
	// Capital greek letters.
	"Alpha", "Gamma", "Delta", "Zeta", "Theta", "Lambda", "Omicron", "Pi",
	"Sigma", "Phi", "Khi", "Psi", "Omega",
	// Common greek letters.
	"alpha", "beta", "gamma", "delta", "epsilon", "zeta", "eta", "theta",
	"lambda", "mu", "nu", "pi", "ro", "varsigma", "sigma", "tau",
	"upsilon", "phi", "khi", "psi", "omega",
	// Mathematical operators.
	"forall", "exists", "nexists", "partial", "emptyset", "in", "notin",
	"mp", "pm", "circ", "infty", "mid", "nmid", "parallel", "nparallel",
	"vee", "wedge", "cup", "cap", "different", "approx", "asequal",
	"equiv", "nequiv", "le", "ge", "subset", "notsubset", "oplus",
	"ominus", "otimes", "oslash", "times", "aleph",
	// Mathematical ensembles.
	"C", "H", "N", "P", "Q", "R", "Z",
	// Arrows.
	"implies", "Leftrightarrow", "mapsto",
	// NULL terminator.
	NULL };

	/*
		Unverified ones ares :
		"different"
		"asequal" = asymptotically equal
		"nequiv"
		"notsubset"
	*/



/*
	Static function prototypes.
*/

// Locates andr references the incoming node.
static reference_t tex_node(const char *ptr, const char *limit);
// Returns the enumeration representation of a reference-defined function name.
static enum Tex_Function tex_function(reference_t ref);
// Extracts the text in a reference, allocates memory and returns copied data.
static char *tex_extract(reference_t ref);

// Resolves (computes width, height and baseline) a node and all its children.
static void tex_resolve_node(struct Tex_Node *node);
// Resolves an expression and all its children nodes.
static void tex_resolve_expr(struct Tex_Expression *expr);
// Computes the height of an opening bracket and its closing pair, if existing.
static void tex_resolve_bracket(struct Tex_Node *bracket);

// Loads a referenced expression into a structure and resolves it.
static struct Tex_Expression *tex_load_expr(reference_t ref);
// Loads a references node into a structure and resolves it.
static struct Tex_Node *tex_load_node(reference_t ref);

// Draws a single node at position (x;y) using the set function.
static void tex_draw_node(struct Tex_Node *node, int x, int y);

// Frees a node.
static void tex_free_node(struct Tex_Node *node);



/**
 *	tex_set_pixel_function()
 *
 *	Sets the function used to draw a pixel.
 *
 *	@param	pixel_function	Pixel drawing function address or symbol.
 */

void tex_set_pixel_function(void (*pixel_function)(int x, int y, int color))
{
	// Setting the pixel drawing function pointer.
	dpixel = pixel_function;
}

/**
 *	tex_set_fonts()
 *
 *	Sets the fonts used for printing. The arguments must be valide Font
 *	structures loaded with libfont appropriated method.
 *
 *	@param	text	Font for displaying text.
 *	@param	sybl	Symbol font.
 */

void tex_set_fonts(struct Font *text, struct Font *sybl)
{
	// Setting the Font structure pointers.
	fonts_text = text;
	fonts_sybl = sybl;
}



/**
 *	tex_element()
 *
 *	Returns a reference on the node that begins at the given address.
 *	The second parameter defines the expression end, in the case the
 *	expression is a reference itself.
 *
 *	@param	ptr	Address of expression.
 *	@param	limit	Expression limit.
 *
 *	@return		A reference_t object on the incoming element.
 */

static reference_t tex_node(const char *ptr, const char *limit)
{
	// Using a pointer reference.
	reference_t ref = { ptr, ptr };
	// Using a counter.
	int c = 0;

	// If element is plain text.
	if(!isfunc(*ptr))
	{
		// Getting the end pointer.
		while(ptr < limit && !isfunc(*ptr)) ptr++;
		// Setting reference value and returning.
		ref.end = ptr;
		return ref;
	}

	// Handling exponents and subscripts.
	if(*ptr == '^' || *ptr == '_')
	{
		// Skipping '^' and '_'.
		ptr++;

		// If no bracket is used, only on character is concerned.
		if(*ptr != '{')
		{
			// ref.end is outside the node, so we increase it.
			ref.end = ptr + 1;
			// Returning the reference.
			return ref;
		}

		// Getting the end bracket.
		while(ptr < limit)
		{
			if(*ptr == '{') c++;
			else if(*ptr == '}') if(!--c) break;
			ptr++;
		}

		// Skipping the end bracket.
		ref.end = ptr + 1;
		return ref;
	}

	// Everything that reaches this point begins with '\\'.
	// Increasing pointer to point at the first command character.
	ptr++;

	// Handling bracket commands.
	if(isbracket(*ptr))
	{
		ref.end = ptr+1;
		return ref;
	}

	// Skipping the command.
	else while(isalpha(*ptr)) ptr++;

	// Going through the argument list by detecting each argument.
	while(ptr < limit && *ptr == '{')
	{
		// Couting arguments, taking nested commands in consideration.
		while(ptr < limit)
		{
			// Counting opening and closing brackets.
			if(*ptr=='{') c++;
			// Argument is closed when c is null.
			else if(*ptr=='}') if(!--c) break;
			// Incrementing ptr.
			ptr++;
		}

		// Incrementing ptr if it has not overflowed.
		if(ptr != limit) ptr++;
	}

	// Setting the reference and returning.
	ref.end = ptr;
	return ref;
}



/**
 *	tex_function()
 *
 *	Returns the Tex_Function enum value of the given literal function.
 *
 *	@param	ref	Function reference.
 *
 *	@return		Corresponding value in Tex_Function enumeration.
 */

enum Tex_Function tex_function(reference_t ref)
{
	const char **f;
	const char *ptr = ref.begin;
	int length = ref.end - ref.begin;

	f = TEX_FUNCTIONS;
	while(*f)
	{
		int len = strlen(*f);
		if(length == len && !strncmp(ptr, *f, length))
			return f - TEX_FUNCTIONS + 1;
		f++;
	}

	f = TEX_SYMBOLS;
	while(*f)
	{
		int len = strlen(*f);
		if(length == len && !strncmp(ptr, *f, length))
			return f - TEX_SYMBOLS + Tex_Symbol;
		f++;
	}

	return 0;
}

/**
 *	tex_extract()
 *
 *	Extracts the text from a reference and returns a copied version as an
 *	allocated char *pointer.
 *
 *	@param	ref		Text reference.
 *
 *	@return			Pointer to copied data.
 */

static char *tex_extract(reference_t ref)
{
	// Calculating length.
	int length = ref.end - ref.begin;
	// Allocating a pointer.
	char *ptr = (char *)malloc(length + 1);
	// Checking alloc failure.
	if(!ptr) return NULL;

	// Copying data and setting a NUL.
	memcpy(ptr, ref.begin, length);
	*(ptr + length) = 0;

	// Returning from function.
	return ptr;
}

/**
 *	tex_resolve_node()
 *
 *	Calculates the width, height and baseline of element depending of its
 *	children information.
 *	Does not use a switch since each block is independent and returns from
 *	the function.
 *
 *	@param	node	Node to resolve.
 */

void tex_resolve_node(struct Tex_Node *node)
{
	struct Tex_Expression **args = node->arguments;

	if(node->function == Tex_Plain)
	{
		node->width = font_length(node->text, fonts_text);
		node->height = fonts_text->height;
		node->line = node->height >> 1;

		return;
	}

	if(node->function == Tex_Symbol)
	{
		node->width = font_length_character(node->character + 32,
			fonts_sybl);
		node->height = fonts_sybl->height;
		node->line = node->height >> 1;

		return;
	}

	if(node->function >= Tex_LeftP && node->function <= Tex_RightB)
	{
		node->width = 3;
		node->height = 7;
		node->line = 3;
		return;
	}

	if(node->function == Tex_Limit)
	{
		int limw = font_length("lim", fonts_text);
		int limh = fonts_text->height;
		node->line = 3;

		if(args[0])
		{
			node->width = max(limw, args[0]->width);
			node->height = limh + args[0]->height + 1;
		}
		else
		{
			node->width = limw;
			node->height = limh;
		}

		return;
	}

	// All nodes after this statement need arguments.
	if(!args[0])
	{
		node->width = 0;
		node->height = 0;
		node->line = 0;

		return;
	}

	if(node->function == Tex_Sup || node->function == Tex_Sub)
	{
		node->width = args[0]->width;
		node->height = args[0]->height;
		node->line = args[0]->line;
		return;
	}

	if(node->function == Tex_Frac)
	{
		node->width = args[0]->width + 2;
		node->height = args[0]->height + 2;
		node->line = args[0]->height + 1;

		if(args[1])
		{
			node->width = max(node->width, args[1]->width) + 2;
			node->height += args[1]->height + 1;
		}
		return;
	}

	if(node->function == Tex_Sqrt)
	{
		node->width = args[0]->width + 6;
		node->height = args[0]->height + 3;
		node->line = args[0]->line + 2;
		return;
	}

	if(node->function == Tex_Vec)
	{
		node->width = args[0]->width;
		node->height = args[0]->height + 4;
		node->line = args[0]->line + 4;
		return;
	}

	if(node->function == Tex_Sum || node->function == Tex_Prod)
	{
		if(!args[1])
		{
			node->width = max(args[0]->width, 8);
			node->height = args[0]->height + 10;
			node->line = 4;
			return;
		}

		int x = max(args[0]->width, args[1]->width);
		node->width = max(x, 8);
		node->height = args[0]->height + args[1]->height + 11;
		node->line = args[0]->height + 5;
		return;
	}

	if(node->function == Tex_Int)
	{
		if(args[1])
		{
			node->width = max(args[0]->width, args[1]->width) + 6;
			node->height = args[0]->height + args[1]->height + 7;
			node->line = args[1]->height + 3;
		}
		else
		{
			node->width = args[0]->width + 6;
			node->height = args[0]->height + 14;
			node->line = 10;
		}
		return;
	}

	node->width = 0;
	node->height = 0;
	node->line = 0;
}

/**
 *	tex_resolve_expr()
 *
 *	This function resolves an expression, that is, calculating the width,
 *	height and baseline of an horizontal list of nodes.
 *
 *	@param	expr	Expression to resolve.
 */

static void tex_resolve_expr(struct Tex_Expression *expr)
{
	// Using a node pointer and pointers to save previous nodes.
	// These previous nodes are used to align exponents and subscripts.
	struct Tex_Node *node = expr->nodes;
	// Using a common integer.
	int n;

	int last_sub = 0, last_line = 0, last_sup_x = 0, last_sub_x = 0;
	// Maximum heights of expression on and under the baseline.
	int sup_height = 0, sub_height = 0;

	// Initializing the default sizes. Not changed if there is no child.
	expr->width = 0;
	expr->height = 0;
	expr->line = 0;

	// Looping through the child linked list.
	while(node)
	{
		// Handling exponents and subscripts.
		if(node->function == Tex_Sup)
		{
			// Calculating new baseline.
			n = last_line + node->height - 3;
			// Setting it if greater.
			if(n > sup_height) sup_height = n;
			// Saving it in last_line.
			last_line = n;

			// Updating last_sup_x.
			last_sup_x += node->width + 1;
		}
		else if(node->function == Tex_Sub)
		{
			// Calculating new height.
			n = last_sub + node->height - 3;
			// Setting it if greater.
			if(n > sub_height) sub_height = n;
			// Saving in last_height.
			last_sub = n;

			// Baseline remains unchanged by adding subscripts.
			// Updating last_sub_x.
			last_sub_x += node->width + 1;
		}
		else
		{
			// Adding the old part of exponents and subscripts.
			expr->width += max(last_sub_x, last_sup_x);
			// Reinitialising these parts.
			last_sup_x = last_sub_x = 0;

			// Adding the node width and spacing.
			expr->width += node->width + 1;

			// Updating maximum height and baseline.
			if(node->line > sup_height) sup_height = node->line;
			last_line = node->line;

			if(node->height - node->line > sub_height)
				sub_height = node->height - node->line;
			last_sub = node->height - node->line;
		}

		// Getting the next node.
		node = node->next;
	}

	// Flushing the width accumulated in exponents and subscripts.
	expr->width += max(last_sub_x, last_sup_x);
	// Removing the last space in the case there is a child.
	if(expr->nodes) expr->width -= 1;
	// Setting the final height and baseline.
	expr->height = sup_height + sub_height;
	expr->line = sup_height;

	// Now looping to resolve the brackets.
	node = expr->nodes;

	// Seeking the opening bracket nodes.
	while(node)
	{
		// Resolving the opening bracket nodes.
		if(node->function == Tex_LeftP || node->function == Tex_LeftS
			|| node->function == Tex_LeftB)
			tex_resolve_bracket(node);

		// Getting the next node.
		node = node->next;
	}
}

void tex_resolve_bracket(struct Tex_Node *node)
{
	struct Tex_Node *tmp = node->next;
	// Using a counter, a last height, and a last line.
	int counter = 1, h = 0, l = 0;
	// Using two heights : one over and one under the baseline.
	int over = 0, under = 0;

	// Looping through the incoming nodes.
	while(tmp)
	{
		if(tmp->function == Tex_Sup)
		{
			l += tmp->height - 3;
			if(l > over) over = l;
		}
		if(tmp->function == Tex_Sub || tmp->function == Tex_Sup)
		{
			h += tmp->height - 3;
			if(h -l > under) under = h - l;

			tmp = tmp->next;
			continue;
		}

		if(tmp->function == node->function) counter++;
		else if(tmp->function == node->function + 1) if(!--counter)
			break;

		// Resetting last height and last line.
		h = tmp->height;
		l = tmp->line;
		// Updating maximum heights.
		if(l > over) over = l;
		if(h - l > under) under = h - l;

		// Selecting the following node.
		tmp = tmp->next;
	}

	// Setting node height and line.
	node->height = over + under;
	node->line = over;

	// Setting up the closing bracket.
	if(tmp)
	{
		// Setting the height.
		tmp->height = node->height;
		// Setting the baseline.
		tmp->line = node->line;
	}
}

/**
 *	tex_load()
 *
 *	Overload for loading expressions. Takes a pointer and defines the limit
 *	as the end of the string (first NUL character).
 *
 *	@param	expression	NULL-terminated expression.
 *
 *	@return			Tex_Expression structure pointer.
 */

struct Tex_Expression *tex_load(const char *expression)
{
	// Getting an end pointer.
	const char *end = expression;
	while(*end) end++;

	// Delegating work to tex_load_pointer().
	return tex_load_pointer(expression, end);
}

/**
 *	tex_load_pointer()
 *
 *	Epxression loading function. Takes two pointers, defining the formula
 *	with a beginning and an end.
 *
 *	@param	expr	Beginning of expression.
 *	@param	limit	End of expression.
 *
 *	@return		Tex_Expression structure pointer.
 */

struct Tex_Expression *tex_load_pointer(const char *expr, const char *limit)
{
	// Constructing a reference.
	reference_t ref = { expr, limit };
	// Parsing the expression.
	struct Tex_Expression *tex = tex_load_expr(ref);

	// Returning the object.
	return tex;
}

/**
 *	tex_load_expr()
 *
 *	Expression loading function.
 *	There is a funny recursion between tex_load_expr() and tex_load_node(),
 *	each one being calling the other until the expression is parsed.
 *	The size of the expression is resolved.
 *
 *	@param	ref		Expression reference.
 *
 *	@return			The first node of a linked node list.
 */

static struct Tex_Expression *tex_load_expr(const reference_t ref)
{
	// Using an expression pointer.
	struct Tex_Expression *expr;
	// Using a generic node and a temporary one.
	struct Tex_Node *node = NULL, *tmp;
	// Using a parser pointer.
	const char *ptr = ref.begin;
	// Using a reference for element selection.
	reference_t element;

	// Checking reference validity.
	if(ref.begin >= ref.end) return NULL;
	// Allocating an expression.
	expr = (struct Tex_Expression *)malloc(sizeof(struct Tex_Expression));
	// Returning if malloc failed.
	if(!expr) return NULL;

	// Looping through the expression reference.
	while(ptr < ref.end)
	{
		// Getting a reference to the next element.
		element = tex_node(ptr, ref.end);
		// Updating parser pointer.
		ptr = element.end;

		// Loading the node.
		tmp = tex_load_node(element);

		// Leaving a link to the next element.
		if(node) node->next = tmp;
		// Else, saving the first node.
		else expr->nodes = tmp;
		// Updating the node address.
		node = tmp;
	}

	// Leaving the last node at the end of the linked list.
	node->next = NULL;

	// Resolving the expression.
	tex_resolve_expr(expr);

	// Returning the first node.
	return expr;
}

/**
 *	tex_load_node()
 *
 *	This function loads a node into a structure from its text reference.
 *	Also calculates the size of the node.
 *
 *	@param	ref		Text reference of the element.
 *
 *	@return			Allocated and loaded node.
 */

static struct Tex_Node *tex_load_node(reference_t ref)
{
	// Allocating a structure.
	struct Tex_Node *node =
		(struct Tex_Node *)malloc(sizeof(struct Tex_Node));
	// Using an expression pointer (argument).
	struct Tex_Expression *arg;
	// Using a parameter counter and a generic integer.
	int counter = 0, x = 0;

	// Setting default value for node->next.
	node->next = NULL;

	// Applying simple formulas in the case of plain text.
	if(!isfunc(*ref.begin))
	{
		// Setting the function.
		node->function = Tex_Plain;
		// Extracting the text.
		node->text = tex_extract(ref);
		// Computing size.
		tex_resolve_node(node);

		// Returning the node.
		return node;
	}

	// Handling exponents and subscript.
	if(*ref.begin == '^' || *ref.begin == '_')
	{
		// Setting the function.
		node->function = (*ref.begin == '^') ? Tex_Sup : Tex_Sub;
		// Skipping the '^' and '_' characters.
		ref.begin++;
		// Excluding the brackets if needed.
		if(*ref.begin == '{') ref.begin++, ref.end--;

		// Loading the argument.
		node->arguments[0] = tex_load_expr(ref);
		node->arguments[1] = NULL;

		tex_resolve_node(node);

		// Returning the node.
		return node;
	}

	// Everything that reaches this point begins with '\\'.
	ref.begin++;
	ref.end = ref.begin;

	// Handling bracket commands.
	if(isbracket(*ref.end))
	{
		ref.end++;
		node->function = tex_function(ref);
		node->arguments[0] = NULL;
		tex_resolve_node(node);
		return node;
	}

	// Getting a reference on the function name.
	while(isalpha(*ref.end)) ref.end++;
	// Getting the enumeration representation.
	node->function = tex_function(ref);

	// Handling symbols.
	if(node->function >= Tex_Symbol)
	{
		// Using the difference to void analyzing the string twice.
		node->character = node->function - Tex_Symbol;
		// Resetting the function identifier to default symbol value.
		node->function = Tex_Symbol;
		// Resolving the node.
		tex_resolve_node(node);
		// Returning the loaded and resolved symbol.
		return node;
	}

	// Moving the reference beginning pointer to the end.
	ref.begin = ref.end;

	// Each iteration reads an argument. Loop is broken when there are no
	// more arguments.
	while(*ref.begin == '{')
	{
		// Getting the argument end, considering nested functions.
		x = 0;
		while(1)
		{
			if(*ref.end == '{') x++;
			else if(*ref.end == '}') if(!--x) break;
			ref.end++;
		}
		// Skipping the argument opening bracket.
		ref.begin++;

		// Loading and resolving the element.
		arg = tex_load_expr(ref);

		// Adding the child to the parameter array.
		node->arguments[counter++] = arg;
		// Skipping the closing bracket.
		ref.begin = ++ref.end;
	}

	// Ending the parameter array.
	node->arguments[counter] = NULL;

	// Resolving the loaded node.
	tex_resolve_node(node);

	// Returning the node.
	return node;
}

void tex_draw(struct Tex_Expression *expr, int x, int y)
{
	struct Tex_Node *node = expr->nodes;
	int last_sup_x = 0, last_sup_y = 0, last_sub_x = 0, last_sub_y = 0;

	while(node)
	{
		if(node->function == Tex_Sup)
		{
			int top_y = last_sup_y - node->height + 3;
			tex_draw_node(node, x + last_sup_x, top_y);

			last_sup_x += node->width + 1;
			last_sup_y = top_y;
		}
		else if(node->function == Tex_Sub)
		{
			int bottom_y = last_sub_y + node->height - 3;
			tex_draw_node(node, x + last_sub_x, last_sub_y - 3);

			last_sub_x += node->width + 1;
			last_sub_y = bottom_y;
		}
		else
		{
			x += max(last_sub_x, last_sup_x);

			tex_draw_node(node, x, y + expr->line - node->line);
			x += node->width + 1;

			last_sup_y = y + expr->line - node->line;
			last_sub_y = last_sup_y + node->height;
			last_sup_x = 0;
			last_sub_x = 0;
		}

		node = node->next;
	}
}

static void tex_draw_node(struct Tex_Node *node, int x, int y)
{
	int w = node->width, h = node->height, l = node->line;
	struct Tex_Expression **args = node->arguments;
	int i;

	if(node->function == Tex_Plain)
	{
		font_text(x, y, node->text, fonts_text);
		return;
	}

	if(node->function == Tex_Symbol)
	{
		char str[2] = { node->character + 32, 0 };
		font_text(x, y, str, fonts_sybl);

		return;
	}

	if(node->function == Tex_Sup || node->function == Tex_Sub)
	{
		tex_draw(args[0], x, y);
	}

	if(node->function == Tex_Frac)
	{
		for(i=0;i<w;i++) (*dpixel)(x+i,y+l,1);
		i = (node->width - args[0]->width) >> 1;
		tex_draw(args[0],x+i,y);
		if(args[1])
		{
			i = (node->width - args[1]->width) >> 1;
			tex_draw(args[1],x+i,y+l+2);
		}
		return;
	}

	if(node->function == Tex_Sqrt)
	{
		for(i=0;i<h;i++) (*dpixel)(x+2,y+i,1);
		for(i=2;i<w;i++) (*dpixel)(x+i,y,1);
		(*dpixel)(x+w-1,y+1,1);
		(*dpixel)(x+w-1,y+2,1);
		(*dpixel)(x+1,y+h-2,1);
		(*dpixel)(x,y+h-3,1);
		tex_draw(args[0],x+4,y+2);
		return;
	}

	if(node->function == Tex_Vec)
	{
		for(i=0 ; i<w ; i++) (*dpixel)(x+i, y+1, 1);
		(*dpixel)(x+w-2, y, 1);
		(*dpixel)(x+w-2, y+2, 1);
		tex_draw(args[0], x, y+4);
	}

	if(node->function == Tex_Sum)
	{
		int n = x + (w >> 1);
		for(i=0;i<8;i++)
			(*dpixel)(n+i-4,y+l-4,1), (*dpixel)(n+i-4,y+l+4,1);
		for(i=0;i<4;i++)
			(*dpixel)(n-i,y+l-i,1), (*dpixel)(n-i,y+l+i,1);
		(*dpixel)(n+3,y+l-3,1);
		(*dpixel)(n+3,y+l+3,1);
	}

	if(node->function == Tex_Prod)
	{
		int n = x + (w >> 1), k = y + l - 4;
		for(i=0;i<8;i++)
		{
			(*dpixel)(n+i-4, k, 1);
			(*dpixel)(n-3, k+i+1, 1);
			(*dpixel)(n+2, k+i+1, 1);
		}
	}

	if(node->function == Tex_Sum || node->function == Tex_Prod)
	{
		int n = (node->width - args[0]->width) >> 1;
		tex_draw(args[0],x+n,y+l+6);
		if(args[1])
		{
			n = (node->width - args[1]->width) >> 1;
			tex_draw(args[1],x+n,y);
		}
		return;
	}

	if(node->function == Tex_Int)
	{
		for(i=1 ; i<h-1 ; i++) (*dpixel)(x+2, y+i, 1);
		(*dpixel)(x+1, y+h-1, 1);
//		(*dpixel)(x, y+h-1, 1);
		(*dpixel)(x, y+h-2, 1);
		(*dpixel)(x+3, y, 1);
//		(*dpixel)(x+4, y, 1);
		(*dpixel)(x+4, y+1, 1);
		tex_draw(args[0], x+6, y+h - args[0]->height);
		if(args[1]) tex_draw(args[1], x+6, y);
		return;
	}

	if(node->function == Tex_Limit)
	{
		int limw = font_length("lim", fonts_text);
		int limh = fonts_text->height;

		int n = (w - limw) >> 1;
		font_text(x+n, y, "lim", fonts_text);
		if(args[0])
		{
			n = (w - args[0]->width) >> 1;
			tex_draw(args[0], x + n, y + limh + 1);
		}
	}

	if(node->function == Tex_LeftP)
	{
		for(i=0 ; i<2 ; i++)
			(*dpixel)(x+2-i, y+i, 1), (*dpixel)(x+2-i, y+h-1-i, 1);
		for(i=2 ; i<h-2 ; i++)
			(*dpixel)(x, y+i, 1);
		return;
	}

	if(node->function == Tex_RightP)
	{
		for(i=0 ; i<2 ; i++)
			(*dpixel)(x+i, y+i, 1), (*dpixel)(x+i, y+h-1-i, 1);
		for(i=2 ; i<h-2 ; i++)
			(*dpixel)(x+2, y+i, 1);
		return;
	}

	if(node->function == Tex_LeftS)
	{
		for(i=0 ; i<2 ; i++)
			(*dpixel)(x+i+1, y, 1), (*dpixel)(x+i+1, y+h-1, 1);
		for(i=0 ; i<h ; i++) (*dpixel)(x, y+i, 1);
	}

	if(node->function == Tex_RightS)
	{
		for(i=0 ; i<2 ; i++)
			(*dpixel)(x+i, y, 1), (*dpixel)(x+i, y+h-1, 1);
		for(i=0 ; i<h ; i++) (*dpixel)(x+2, y+i, 1);
	}

	if(node->function == Tex_LeftB)
	{
		int n = (h-1) >> 1;
		for(i=1 ; i<h-1 ; i++) if(i!=n) (*dpixel)(x+1, y+i, 1);
		(*dpixel)(x, y+n, 1);
		(*dpixel)(x+2, y, 1);
		(*dpixel)(x+2, y+h-1, 1);
	}

	if(node->function == Tex_RightB)
	{
		int n = (h-1) >> 1;
		for(i=1 ; i<h-1 ; i++) if(i!=n) (*dpixel)(x+1, y+i, 1);
		(*dpixel)(x+2, y+n, 1);
		(*dpixel)(x, y, 1);
		(*dpixel)(x, y+h-1, 1);
	}
}

/**
 *	tex_free()
 *
 *	Frees a loaded expression. Expressions should not be used in any tex
 *	function once freed.
 *
 *	@param	expr	Expression to free.
 */

void tex_free(struct Tex_Expression *expr)
{
	// Getting the first node.
	struct Tex_Node *node = expr->nodes, *tmp;

	// Freeing all the child nodes.
	while(node)
	{
		// Saving the next node.
		tmp = node->next;
		// Freeing the current node.
		tex_free_node(node);
		// Selecting the next node.
		node = tmp;
	}

	// Freeing the Tex_Expression structure.
	free(expr);
}

/**
 *	tex_free_node()
 *
 *	Frees a node and all its arguments.
 *
 *	@param	node	Node to free.
 */

static void tex_free_node(struct Tex_Node *node)
{
	// Symbols have nothing dynamically allocated.
	if(node->function == Tex_Symbol) return;

	// Freeing plain text nodes.
	if(node->function == Tex_Plain)
	{
		free(node->text);
	}

	// Freeing function nodes.
	else
	{
		// Selecting the arguments.
		struct Tex_Expression **args = node->arguments;
		// Freeing the arguments.
		while(*args) tex_free(*args), args++;
	}

	// Freeing the Tex_Node structure.
	free(node);
}

/**
 *	tex_debug()
 *
 *	Debugs a linked list on the given stream.
 *
 *	@param	node	First node to debug.
 *	@param	output	Stream to output to.
 *	@param	level	Level of debugging.
 */

/*

static void tex_debug_output(struct Tex_Node *node, FILE *output, int level,
	int mark);

void tex_debug(struct Tex_Expression *expr, FILE *output, int level)
{
	struct Tex_Node *node = expr->nodes;
	int i;

	while(node)
	{
		tex_debug_output(node, output, level, node==expr->nodes);

		if(node->function != Tex_Plain) for(i=0;node->arguments[i];i++)
			tex_debug(node->arguments[i],output,level+1);

		node = node->next;
	}
}

static void tex_debug_output(struct Tex_Node *node, FILE *output, int level,
	int mark)
{
	int i;
	for(i=0;i<level;i++) fputc('\t',output);
	fputc(mark ? '-' : ' ', output);
	fputc(' ', output);

	if(node->function == Tex_Symbol)
	{
		fputc('{', output);
		fputs(TEX_SYMBOLS[node->character], output);
		fputc('}', output);
	}
	else if(node->function != Tex_Plain)
	{
		fputc('[', output);
		fputs(TEX_FUNCTIONS[node->function-1], output);
		fputc(']', output);
	}
	else
	{
		fputc('`', output);
		fputs(node->text, output);
		fputc('`', output);
	}
	fputc('\n', output);
}

*/
