#ifndef _TEX_0_9_H
	#define _TEX_0_9_H

/*
	Headers inclusion.
*/

/*
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
// #include <ctype.h>
*/

#include <fxlib.h>
#include <stdref.h>
#include "libfont-3.1.h"



/*
	Macro constants.
*/

#define TEX_ARGUMENTS 2



/*
	Composed types definitions.
*/

/**
 *	Tex_Function
 *
 *	This enumeration represents a function. Every node can be seen as plain
 *	text, function, or symbol. The Tex_Function enumeration allows to
 *	handle intuitively these various functions.
 *
 *	Tex_Symbol is at the end because the static function tex_function(),
 *	which returns the enumeration representation of a function name,
 *	returns Tex_Symbol + character_index when a symbol name is detected.
 */

enum Tex_Function
{
	// Plain text elements (no function).
	Tex_Plain = 0,

	/*
		Various functions.
	*/

	Tex_Frac,
	Tex_Sqrt,
	Tex_Vec,

	Tex_Sum,
	Tex_Prod,

	Tex_Int,
	Tex_Limit,

	Tex_Sup,
	Tex_Sub,

	Tex_LeftP,
	Tex_RightP,
	Tex_LeftS,
	Tex_RightS,
	Tex_LeftB,
	Tex_RightB,

	// Symbols.
	Tex_Symbol
};

/**
 *	Tex_Expression
 *
 *	An expression is basically an horizontal list of nodes. A node can be
 *	plain text, a function or a symbol. An expression is just a container
 *	handling the list of the nodes and their overall size.
 */

struct Tex_Expression
{
	// Expression dimensions and geometry.
	unsigned width	:16;
	unsigned height	:16;
	unsigned line	:16;

	// Top of node linked list.
	struct Tex_Node *nodes;
};

/**
 *	Tex_Node
 *
 *	This is the main structure in the program. A node is basically what is
 *	loaded from string and drawn on the string. Since expressions handle
 *	the node flow, all a node has to focus on is its own text, function, or
 *	symbol. Actually, that's not exactly true, nodes are handled as linked
 *	lists, so there's a pointer to the next element.
 *
 *	Depending on its nature, the main data can be stored is a dynamically-
 *	allocated string, in an array of arguments, or in a single character.
 *	The number of argument pointers is fixed, but the arguments may have to
 *	be dynamically linked if functions with variable arguments come to
 *	appear (particularly matrices).
 *
 *	A node also has its own width, height and baseline. The baseline of a
 *	node or expression is the vertical position, beginning from 0 at the
 *	top of the element, that indicates where the middle line is.
 *	Take for instance, \frac{a}{\frac{b}{c}} and \frac{\frac{a}{b}}{c}. In
 *	the first case, the baseline is under `a`, where in the seconde case,
 *	it is located under `b'. It is used for vertical alignment purposes.
 */

struct Tex_Node
{
	// Node function.
	enum Tex_Function function;

	// Node parameters.
	union
	{
		// Plain text content.
		char *text;
		// Function arguments.
		struct Tex_Expression *arguments[TEX_ARGUMENTS + 1];
		// Symbol identifier.
		int character;
	};
	// Linked list pointer, on next element.
	struct Tex_Node *next;

	// Node dimensions and geometry.
	unsigned width	:16;
	unsigned height	:16;
	unsigned line	:16;
};



/*
	Function prototypes.
*/

// Sets the pixel function.
void tex_set_pixel_function(void (*pixel_function)(int x, int y, int color));
// Sets the fonts.
void tex_set_fonts(struct Font *text_font, struct Font *sybl_font);
// Loading function.
struct Tex_Expression *tex_load(const char *expression);
// Loading function, using pointers.
struct Tex_Expression *tex_load_pointer(const char *begin, const char *end);
// Draws an expression (logically named `tex_draw_expr()`).
void tex_draw(struct Tex_Expression *expression, int x, int y);
// Frees an expression.
void tex_free(struct Tex_Expression *expr);
// Debugs a loaded expression.
// void tex_debug(struct Tex_Expression *expr, FILE *output, int level);



#endif // _TEX_0_9_H
